""" Widget that shows the content of project in a electronic labnotebook """
import logging
from enum import Enum
from typing import Optional, Any
from PySide6.QtWidgets import QLabel, QVBoxLayout, QWidget, QMenu, QMessageBox, QTextEdit, QScrollArea # pylint: disable=no-name-in-module
from PySide6.QtGui import QStandardItemModel, QStandardItem, QAction                                   # pylint: disable=no-name-in-module
from PySide6.QtCore import Slot, Qt, QItemSelectionModel, QModelIndex                                  # pylint: disable=no-name-in-module
from anytree import PreOrderIter, Node
from .projectTreeView import TreeView
from ..guiStyle import TextButton, Action, Label, showMessage, widgetAndLayout, getColor
from ..miscTools import createDirName, markdownStyler
from ..guiCommunicate import Communicate

class Project(QWidget):
  """ Widget that shows the content of project in a electronic labnotebook """
  def __init__(self, comm:Communicate):
    super().__init__()
    self.comm = comm
    comm.changeProject.connect(self.change)
    self.mainL = QVBoxLayout()
    self.setLayout(self.mainL)
    self.tree:Optional[TreeView]             = None
    self.model:Optional[QStandardItemModel]  = None
    self.infoWSA:Optional[QWidget]           = None
    self.infoW_:Optional[QWidget]            = None
    self.commentTE:Optional[QWidget]         = None
    self.actHideDetail = QAction()
    self.actionHideItems   = QAction()
    self.actionHideProject = QAction()
    self.actionFoldAll     = QAction()
    self.projID = ''
    self.taskID = ''
    self.docProj:dict[str,Any]= {}
    self.showAll= True
    self.showDetailsAll = False
    self.btnAddSubfolder:Optional[TextButton] = None
    self.lineSep = 20
    self.countLines = -1


  def projHeader(self) -> None:
    """
    Initialize / Create header of page
    """
    self.docProj = self.comm.backend.db.getDoc(self.projID)
    # remove if still there
    for i in reversed(range(self.mainL.count())): #remove old
      self.mainL.itemAt(i).widget().setParent(None)  # type: ignore
    logging.debug('ProjectView elements at 2: %i',self.mainL.count())
    # TOP LINE includes name on left, buttons on right
    _, topLineL       = widgetAndLayout('H',self.mainL,'m')
    hidden, menuTextHidden = ('     \U0001F441', 'Mark project as shown') \
                       if [b for b in self.docProj['-branch'] if False in b['show']] else \
                       ('', 'Mark project as hidden')
    topLineL.addWidget(Label(self.docProj['-name']+hidden, 'h2'))
    showStatus = '(Show all items)' if self.showAll else '(Hide hidden items)'
    topLineL.addWidget(QLabel(showStatus))
    topLineL.addStretch(1)
    # buttons in top line
    buttonW, buttonL = widgetAndLayout('H', spacing='m')
    topLineL.addWidget(buttonW, alignment=Qt.AlignTop)  # type: ignore
    self.btnAddSubfolder = TextButton('Add subfolder', self, [Command.ADD_CHILD], buttonL)
    TextButton('Edit project',                         self, [Command.EDIT],      buttonL)
    visibility = TextButton(          'Visibility',    self, [],                  buttonL)
    visibilityMenu = QMenu(self)
    self.actHideDetail = Action('Hide project details',self, [Command.SHOW_PROJ_DETAILS],visibilityMenu)
    menuTextItems = 'Hide hidden items' if self.showAll else 'Show hidden items'
    minimizeItems = 'Show all item details' if self.showDetailsAll else 'Hide all item details'
    self.actionHideItems   = Action( menuTextItems,    self, [Command.HIDE_SHOW_ITEMS],  visibilityMenu)
    self.actionHideProject = Action( menuTextHidden,   self, [Command.HIDE],             visibilityMenu)
    self.actionFoldAll     = Action( minimizeItems,    self, [Command.SHOW_DETAILS],   visibilityMenu)
    visibility.setMenu(visibilityMenu)
    more = TextButton('More',           self, [], buttonL)
    moreMenu = QMenu(self)
    Action('Scan',                      self, [Command.SCAN], moreMenu)
    for doctype in self.comm.backend.db.dataLabels:
      if doctype[0]!='x':
        icon = self.comm.backend.db.dataHierarchy[doctype].get('icon','')
        icon = 'fa.asterisk' if icon=='' else icon
        Action(f'table of {doctype}',   self, [Command.SHOW_TABLE, doctype], moreMenu, icon=icon)
    Action('table of unidentified',     self, [Command.SHOW_TABLE, '-'],     moreMenu, icon='fa5.file')
    moreMenu.addSeparator()
    Action('Delete project',            self, [Command.DELETE], moreMenu)
    more.setMenu(moreMenu)

    # Details section
    self.infoWSA = QScrollArea()
    self.infoWSA.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
    self.infoWSA.setWidgetResizable(True)
    self.infoW_, infoL = widgetAndLayout('V')
    self.infoWSA.setWidget(self.infoW_)
    self.mainL.addWidget(self.infoWSA)
    if not self.docProj['-gui'][0]:
      self.infoWSA.hide()
      self.actHideDetail.setText('Show project details')
    # details
    tags = ', '.join([f'#{i}' for i in self.docProj['-tags']]) if '-tags' in self.docProj else ''
    infoL.addWidget(QLabel(f'Tags: {tags}'))
    self.countLines = 0
    for key,value in self.docProj.items():
      if key[0] in {'_','-'} or 'from ' in key or key in {'comment'}:
        continue
      labelW = QLabel(f'{key.title()}: {str(value)}')
      infoL.addWidget(labelW)
      self.countLines += 1
    # comment
    commentW, commentL   = widgetAndLayout('H', infoL, 's')
    commentW.resizeEvent = self.commentResize # type: ignore
    labelW = QLabel('Comment:')
    # labelW.setStyleSheet('padding-top: 5px') #make "Comment:" text aligned with other content, not with text-edit
    commentL.addWidget(labelW, alignment=Qt.AlignTop)   # type: ignore[call-arg]
    self.commentTE = QTextEdit()
    self.commentTE.setMarkdown(markdownStyler(self.docProj.get('comment', '')))
    bgColor = getColor(self.comm.backend, 'secondaryDark')
    fgColor = getColor(self.comm.backend, 'primaryText')
    self.commentTE.setStyleSheet(f"border: none; padding: 0px; background-color: {bgColor}; color: {fgColor}")
    self.commentTE.setReadOnly(True)
    self.commentResize(None)
    commentL.addWidget(self.commentTE)
    return


  def commentResize(self, _:Any) -> None:
    """ called if comment is resized because widget initially/finally knows its size
    - comment widget is hard coded size it depends on the rendered size
    """
    if self.commentTE is None or self.infoWSA is None or self.infoW_ is None:
      return
    self.commentTE.document().setTextWidth(self.infoWSA.width())
    height:int = self.commentTE.document().size().toTuple()[1]
    self.infoW_.setMaximumHeight(height + (self.countLines+1)*self.lineSep     -12)
    self.infoWSA.setMaximumHeight(height + (self.countLines+1)*self.lineSep  -10)
    return


  @Slot(str, str)
  def change(self, projID:str, docID:str) -> None:
    """
    What happens when user clicks to change project

    Args:
      projID (str): document id of project; if empty, just refresh
      docID (str): document id of focus item, if not given focus at project
    """
    logging.debug('project:changeProject |%s|%s|',projID,docID)
    #initialize
    for i in reversed(range(self.mainL.count())): #remove old
      self.mainL.itemAt(i).widget().setParent(None)  # type: ignore
    logging.debug('ProjectView elements at 1: %i',self.mainL.count())
    if projID!='':
      self.projID         = projID
      self.taskID         = docID
      self.comm.projectID = projID
    selectedIndex = None
    self.model = QStandardItemModel()
    self.tree = TreeView(self, self.comm, self.model)
    # self.tree.setSelectionBehavior(QAbstractItemView.SelectRows)
    # self.tree.setSelectionMode(QAbstractItemView.SingleSelection)
    self.model.itemChanged.connect(self.modelChanged)
    rootItem = self.model.invisibleRootItem()
    #Populate model body of change project: start recursion
    nodeHier = self.comm.backend.db.getHierarchy(self.projID, allItems=self.showAll)
    for node in PreOrderIter(nodeHier, maxlevel=2):
      if node.is_root:         #Project header
        self.projHeader()
      else:
        rootItem.appendRow(self.iterateTree(node))
    # collapse / expand depending on stored value
    # by iterating each leaf, and converting item and index
    root = self.model.invisibleRootItem()
    self.setExpandedState(root)
    if selectedIndex is not None:
      self.tree.selectionModel().select(selectedIndex, QItemSelectionModel.Select)
      self.tree.setCurrentIndex(selectedIndex)# Item(selectedItem)
    self.mainL.addWidget(self.tree)
    logging.debug('ProjectView elements at 4: %i',self.mainL.count())
    if len(nodeHier.children)>0 and self.btnAddSubfolder is not None:
      self.btnAddSubfolder.setVisible(False)
    self.tree.expanded.connect(lambda index: self.actionExpandCollapse(index, True))
    self.tree.collapsed.connect(lambda index: self.actionExpandCollapse(index, False))
    return


  def setExpandedState(self, node:QStandardItem) -> None:
    """ Recursive function to set the expanded state of nodes

    Args:
      node (QStandardItem): node to process
    """
    if self.model is None or self.tree is None:
      return
    for iRow in range(node.rowCount()):
      item = node.child(iRow)
      data = item.data(role=Qt.UserRole+1)         # type: ignore[operator]
      if data['hierStack'].split('/')[-1][0]=='x':
        index = self.model.indexFromItem(item)
        self.tree.setExpanded(index, data['gui'][1])
      self.setExpandedState(item)
    return


  def actionExpandCollapse(self, index:QModelIndex, flag:bool) -> None:
    """ Action upon expansion or collapsing of folder (showing its children)

    Args:
      index (QModelIndex): index that send the signal
      flag (bool): true=expand=show-children, false=collapse=hide-children
    """
    if self.model is None:
      return
    gui  = [index.data(Qt.UserRole+1)['gui'][0]]+[flag]                                   # type: ignore[operator]
    docID = index.data(Qt.UserRole+1)['hierStack'].split('/')[-1]                         # type: ignore[operator]
    self.model.itemFromIndex(index).setData({ **index.data(Qt.UserRole+1), **{'gui':gui}})# type: ignore[operator]
    self.comm.backend.db.setGUI(docID, gui)
    return


  def execute(self, command:list[Any]) -> None:
    """
    Event if user clicks button in the center

    Args:
      command (list): list of commands
    """
    if command[0] is Command.EDIT:
      self.comm.formDoc.emit(self.docProj)
      self.change(self.projID,'')
      #collect information and then change
      oldPath = self.comm.backend.basePath/self.docProj['-branch'][0]['path']
      if oldPath.is_dir():
        newPath = self.comm.backend.basePath/createDirName(self.docProj['-name'],'x0',0)
        oldPath.rename(newPath)
      self.comm.changeSidebar.emit('redraw')
    elif command[0] is Command.DELETE:
      ret = QMessageBox.critical(self, 'Warning', 'Are you sure you want to delete project?', \
                      QMessageBox.StandardButton.No | QMessageBox.StandardButton.Yes,  # type: ignore[operator]
                      QMessageBox.StandardButton.No)
      if ret==QMessageBox.StandardButton.Yes:
        #delete database and rename folder
        doc = self.comm.backend.db.remove(self.projID)
        if '-branch' in doc and len(doc['-branch'])>0 and 'path' in doc['-branch'][0]:
          oldPath = self.comm.backend.basePath/doc['-branch'][0]['path']
          newPath = self.comm.backend.basePath/('trash_'+doc['-branch'][0]['path'])
          nextIteration = 1
          while newPath.is_dir():
            newPath = self.comm.backend.basePath/(f"trash_{doc['-branch'][0]['path']}_{nextIteration}")
            nextIteration += 1
          oldPath.rename(newPath)
        # go through children, remove from DB
        children = self.comm.backend.db.getView('viewHierarchy/viewHierarchy', startKey=self.projID)
        for line in children:
          self.comm.backend.db.remove(line['id'])
        #update sidebar, show projects
        self.comm.changeSidebar.emit('redraw')
        self.comm.changeTable.emit('x0','')
    elif command[0] is Command.SCAN:
      self.comm.backend.scanProject(self.comm.progressBar, self.projID, self.docProj['-branch'][0]['path'])
      self.comm.changeSidebar.emit('redraw')
      showMessage(self, 'Information','Scanning finished')
    elif command[0] is Command.SHOW_PROJ_DETAILS:
      self.docProj['-gui'][0] = not self.docProj['-gui'][0]
      self.comm.backend.db.setGUI(self.projID, self.docProj['-gui'])
      if self.infoWSA is not None and self.infoWSA.isHidden():
        self.infoWSA.show()
        self.actHideDetail.setText('Hide project details')
      elif self.infoWSA is not None:
        self.infoWSA.hide()
        self.actHideDetail.setText('Show project details')
    elif command[0] is Command.HIDE:
      self.comm.backend.db.hideShow(self.projID)
      self.docProj = self.comm.backend.db.getDoc(self.projID)
      if [b for b in self.docProj['-branch'] if False in b['show']]: # hidden->go back to project table
        self.comm.changeSidebar.emit('')
        self.comm.changeTable.emit('x0','') # go back to project table
      else:
        self.change('', '')
        self.comm.changeSidebar.emit('')
    elif command[0] is Command.SHOW_DETAILS and self.tree is not None:
      def recursiveRowIteration(index:QModelIndex) -> None:
        for subRow in range(self.tree.model().rowCount(index)):   # type: ignore[union-attr]
          subIndex = self.tree.model().index(subRow,0, index)     # type: ignore[union-attr]
          subItem  = self.tree.model().itemFromIndex(subIndex)    # type: ignore[union-attr]
          docID    = subItem.data()['hierStack'].split('/')[-1]
          gui      = subItem.data()['gui']
          gui[0]   = self.showDetailsAll
          subItem.setData({ **subItem.data(), **{'gui':gui}})
          self.comm.backend.db.setGUI(docID, gui)
          recursiveRowIteration(subIndex)
      recursiveRowIteration(self.tree.model().index(-1,0))
      self.showDetailsAll = not self.showDetailsAll
      if self.showDetailsAll:
        self.actionFoldAll.setText('Show all item details')
      else:
        self.actionFoldAll.setText('Hide all item details')
    elif command[0] is Command.HIDE_SHOW_ITEMS:
      self.showAll = not self.showAll
      self.change('','')
    elif command[0] is Command.ADD_CHILD:
      self.comm.backend.cwd = self.comm.backend.basePath/self.docProj['-branch'][0]['path']
      title = self.comm.backend.db.dataHierarchy['x1']['title'].lower()[:-1]
      self.comm.backend.addData('x1', {'-name':f'new {title}'}, [self.projID])
      self.change('','') #refresh project
    elif command[0] is Command.SHOW_TABLE:
      self.comm.changeTable.emit(command[1], self.projID)
    else:
      print("**ERROR project menu unknown:",command)
    return


  def modelChanged(self, item:QStandardItem) -> None:
    """
    Autocalled after drag-drop and other changes, record changes to backend and database directly

    Args:
      item (QStandardItem): item changed, new location
    """
    #gather old information
    db       = self.comm.backend.db
    if not item.data():
      return
    stackOld = item.data()['hierStack'].split('/')[:-1]
    docID    = item.data()['hierStack'].split('/')[-1]
    doc      = db.getDoc(docID)
    if '-branch' not in doc or not stackOld: #skip everything if project or not contain branch
      return
    branchOldList= [i for i in doc['-branch'] if i['stack']==stackOld]
    if len(branchOldList)!=1:
      self.change('','')
      return
    branchOld = branchOldList[0]
    childOld = branchOld['child']
    branchIdx= doc['-branch'].index(branchOld)
    siblingsOld = db.getView('viewHierarchy/viewHierarchy', startKey=' '.join(stackOld))
    siblingsOld = [i for i in siblingsOld if len(i['key'].split(' '))==len(stackOld)+1 and \
                                                  i['value'][0]>branchOld['child'] and i['value'][0]<9999]
    #gather new information
    stackNew = []  #create reversed
    currentItem = item
    while currentItem.parent() is not None:
      currentItem = currentItem.parent()
      docIDj = currentItem.data()['hierStack'].split('/')[-1]
      stackNew.append(docIDj)
    stackNew = [self.projID] + stackNew[::-1]  #add project id and reverse
    childNew = item.row()
    if branchOld['path'] is not None and not branchOld['path'].startswith('http'):
      dirNameNew= createDirName(doc['-name'],doc['-type'][0],childNew) # determine path: do not create yet
      parentDir = db.getDoc(stackNew[-1])['-branch'][0]['path']
      pathNew = f'{parentDir}/{dirNameNew}'
    else:
      pathNew = branchOld['path']
    siblingsNew = db.getView('viewHierarchy/viewHierarchy', startKey=' '.join(stackNew))
    siblingsNew = [i for i in siblingsNew if len(i['key'].split(' '))==len(stackNew)+1 and \
                                                   i['value'][0]>=childNew and i['value'][0]<9999]
    logging.debug('Change project: docID %s branch %i | old stack %s child %i | new stack %s child %i path %s'\
                  , docID, branchIdx, str(stackOld), childOld, str(stackNew), childNew, pathNew)
    if stackOld==stackNew and childOld==childNew:  #nothing changed, just redraw
      return
    # change item in question
    db.updateBranch(docID=docID, branch=branchIdx, stack=stackNew, path=pathNew, child=childNew)
    item.setData(item.data() | {'hierStack': '/'.join(stackNew+[docID])})
    # change siblings
    for line in siblingsOld:
      db.updateBranch(  docID=line['id'], branch=line['value'][4], child=line['value'][0]-1)
    for line in siblingsNew:
      if line['id']!=docID:
        db.updateBranch(docID=line['id'], branch=line['value'][4], child=line['value'][0]+1)
    return


  def iterateTree(self, nodeHier:Node) -> QStandardItem:
    """
    Recursive function to translate the hierarchical node into a tree-node

    Args:
      nodeHier (Anytree.Node): anytree node

    Returns:
      QtTreeWidgetItem: tree node
    """
    #prefill docID
    hierStack = '/'.join([i.id for i in nodeHier.ancestors]+[nodeHier.id])
    gui = nodeHier.gui
    nodeTree = QStandardItem(nodeHier.name)
    nodeTree.setData({"hierStack":hierStack, "docType":nodeHier.docType, "gui":gui})
    if nodeHier.id[0]=='x':
      nodeTree.setFlags(Qt.ItemIsEnabled | Qt.ItemIsSelectable | Qt.ItemIsDragEnabled | Qt.ItemIsDropEnabled) # type: ignore
    else:
      nodeTree.setFlags(Qt.ItemIsEnabled | Qt.ItemIsSelectable | Qt.ItemIsDragEnabled) # type: ignore
    children = []
    for childHier in nodeHier.children:
      childTree = self.iterateTree(childHier)
      children.append(childTree)
    if children:
      nodeTree.appendRows(children)
    return nodeTree


class Command(Enum):
  """ Commands used in this file """
  EDIT   = 1
  DELETE = 2
  SCAN   = 3
  HIDE   = 4
  SHOW_PROJ_DETAILS = 5
  HIDE_SHOW_ITEMS    = 6
  SHOW_DETAILS     = 7
  ADD_CHILD          = 8
  SHOW_TABLE         = 9



""" Represents a data upload task. """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: upload_task_thread.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.
import datetime
import random
from time import sleep

from PySide6 import QtCore

from pasta_eln.GUI.dataverse.upload_widget_base import Ui_UploadWidgetFrame
from pasta_eln.dataverse.config_model import ConfigModel
from pasta_eln.dataverse.database_api import DatabaseAPI
from pasta_eln.dataverse.generic_task_object import GenericTaskObject
from pasta_eln.dataverse.upload_model import UploadModel
from pasta_eln.dataverse.upload_status_values import UploadStatusValues
from pasta_eln.dataverse.utils import update_status


class DataUploadTask(GenericTaskObject):
  """
  Represents a data upload task.

  Explanation:
      This class handles the data upload task, including tracking progress, updating status,
      and managing the upload model.
      It provides methods to start the task, cancel the task, and handle the initialization of the task.

  Args:
      widget (Ui_UploadWidgetFrame): The widget containing the upload task elements.

  Returns:
      None
  """
  progressChanged = QtCore.Signal(int)
  statusChanged = QtCore.Signal(str)
  uploadModelCreated = QtCore.Signal(str)

  def __init__(self, widget: Ui_UploadWidgetFrame) -> None:
    """
    Initializes a DataUploadTask instance.

    Explanation:
        This method initializes a DataUploadTask instance by connecting signals to the widget elements,
        setting up attributes for project name, database API, upload model, and configuration model,
        and connecting the cancel button to the cancel signal.

    Args:
        widget (Ui_UploadWidgetFrame): The widget containing the upload task elements.

    Returns:
        None
    """
    super().__init__()
    self.progressChanged.connect(widget.uploadProgressBar.setValue)
    self.statusChanged.connect(lambda status: update_status(status, widget.statusIconLabel, widget.statusLabel))
    self.project_name = widget.uploadProjectLabel.text()
    self.db_api = DatabaseAPI()
    self.upload_model = UploadModel(project_name=self.project_name,
                                    status=UploadStatusValues.Queued.name,
                                    log=f"Upload initiated for project {self.project_name} at {datetime.datetime.now().isoformat()}\n")
    self.upload_model = self.db_api.create_model_document(self.upload_model)  # type: ignore[assignment]
    self.config_model = self.db_api.get_model(self.db_api.config_doc_id, ConfigModel)
    widget.uploadCancelPushButton.clicked.connect(self.cancel.emit)

  def start_task(self) -> None:
    """
    Starts the data upload task.

    Explanation:
        This method starts the data upload task by emitting signals to update the progress and status,
        creating the upload model, and performing the upload process.
        It updates the progress and status, and updates the upload model accordingly.
        It also handles the cancellation of the task and emits the appropriate signals.
        It also handles the upload tasks to dataverse
          - Generating ELN file
          - Uploading to dataverse
          - Saving the upload model whenever needed

    """
    super().start_task()
    self.progressChanged.emit(0)
    self.statusChanged.emit(UploadStatusValues.Queued.name)
    self.statusChanged.emit(UploadStatusValues.Uploading.name)
    self.uploadModelCreated.emit(self.upload_model.id)
    self.upload_model.log = "Generating ELN file....."
    for progressbar_value in range(101):
      self.progressChanged.emit(progressbar_value)
      if self.cancelled:
        self.upload_model.status = UploadStatusValues.Cancelled.name
        break
      self.upload_model.log = f"Uploading.......... Progress: {progressbar_value}%"
      self.upload_model.status = UploadStatusValues.Uploading.name
      self.db_api.update_model_document(self.upload_model)

      sleep(random.uniform(0.01, 0.06))
    if not self.cancelled:
      self.upload_model.log = f"Uploading completed at {datetime.datetime.now().isoformat()}"
      self.upload_model.finished_date_time = datetime.datetime.now().isoformat()
      self.upload_model.status = UploadStatusValues.Finished.name
      self.upload_model.dataverse_url = f"https://dataverse.harvard.edu/dataverse/{self.upload_model.project_name}"
      self.upload_model.log = f"Uploading to url: {self.upload_model.dataverse_url}"
    self.db_api.update_model_document(self.upload_model)
    self.finished.emit()
    self.statusChanged.emit(UploadStatusValues.Cancelled.name if self.cancelled else UploadStatusValues.Finished.name)

  def cancel_task(self) -> None:
    """
    Cancels the data upload task.

    Explanation:
        This method cancels the data upload task by calling the super-class's cancel_task method.
        It updates the upload model log and status accordingly, and emits the statusChanged signal.

    """
    super().cancel_task()
    self.upload_model.log = f"Cancelled at {datetime.datetime.now().isoformat()}"
    self.upload_model.status = UploadStatusValues.Cancelled.name
    self.statusChanged.emit(UploadStatusValues.Cancelled.name)



""" Class for interaction with couchDB """
import traceback, logging, time, json, os, re, base64, io
from typing import Any, Optional, Union
from pathlib import Path
from anytree import Node
from anytree.search import find_by_attr
from PIL import Image
from cloudant.client import CouchDB
from cloudant.replicator import Replicator
from .fixedStringsJson import defaultDataHierarchy, defaultDataHierarchyNode
from .handleDictionaries import dataHierarchy_pre_to_V4
from .miscTools import tracebackString, outputString

class Database:
  """
  Class for interaction with couchDB
  """
  def __init__(self, user:str, password:str, databaseName:str, configuration:dict[str,Any],
               resetDataHierarchy:bool=False, basePath:Path=Path()):
    """
    Args:
      user (string): user name to local database
      password (string): password to local database
      databaseName (string): local database name
      configuration (dict): configuration of GUI elements
      resetDataHierarchy (bool): reset dataHierarchy
    """
    try:
      self.client = CouchDB(user, password, url='http://127.0.0.1:5984', connect=True)
    except Exception:
      print('**ERROR dit01: Could not connect with username+password to local server')
      raise  #there is no possibility to recover
    self.databaseName = databaseName
    if self.databaseName in self.client.all_dbs():
      self.db = self.client[self.databaseName]
    else:
      self.db = self.client.create_database(self.databaseName)  #tests and initial creation of example data set requires this
    # check if default documents exist and create
    if '-dataHierarchy-' not in self.db or resetDataHierarchy:
      if '-dataHierarchy-' in self.db:
        print('Info: remove old dataHierarchy')
        self.db['-dataHierarchy-'].delete()
      self.dataHierarchy = defaultDataHierarchy
      self.db.create_document(self.dataHierarchy)
      self.initDocTypeViews( configuration['tableColumnsMax'] )
      self.initGeneralViews()
    self.dataHierarchy = dict(self.db['-dataHierarchy-'])
    if '-version' in self.dataHierarchy and self.dataHierarchy['-version'] < 4:
      logging.info('Convert data hierarchy to V4.0')
      dataHierarchy_pre_to_V4(self.dataHierarchy)
      if '-dataHierarchy-' in self.db:
        self.db['-dataHierarchy-'].delete()
      self.db.create_document(self.dataHierarchy)
      if '-ontology-' in self.db:
        self.db['-ontology-'].delete()
    # temporary changes for version 2.5: remove afterwards: code does not harm but would be legacy then
    testDocIDs = [doc['_id'] for doc in self.db if doc['_id'].startswith('x-')]
    if testDocIDs and '-gui' not in self.db[testDocIDs[0]]:
      for doc in self.db:
        if doc['_id'].startswith('_') or doc['_id'].endswith('-'):
          continue
        if '-gui' not in doc:
          doc['-gui'] = [True, True]
          doc.save()
    if '-version' not in self.dataHierarchy or self.dataHierarchy['-version']!=4:
      print(F"**ERROR wrong dataHierarchy version: {self.dataHierarchy['-version']}")
      raise ValueError(f"Wrong dataHierarchy version {self.dataHierarchy['-version']}")
    self.dataLabels = {k:v['title'] for k,v in self.dataHierarchy.items() if k[0] not in ['_','-']}
    self.basePath   = basePath
    return


  def initDocTypeViews(self, tableColumnsMax:int, docTypeChange:str='', columnsChange:list[str]=[]) -> None:
    """
    for the individual docTypes

    Args:
      tableColumnsMax (int): max. number of columns in the docType tables
      docTypeChange (str): if change columns for docType: give docType
      columnsChange (list): change table / view columns to these
    """
    tracebackString(True, 'initDocTypeView')
    # print('SB: initDocViews newColumns',newColumns)
    oldColumnNames = self.getColumnNames()
    jsDefault = 'if ($docType$) {doc["-branch"].forEach(function(branch){emit($key$, [$outputList$]);});}'
    viewCode = {}
    normalDocTypes = [i for i in self.dataHierarchy if i[0] not in ['_','-']]
    for docType in normalDocTypes+['-']:
      if docType == 'x0':
        newString = "doc['-type']=='x0' && (doc['-branch'][0].show.every(function(i) {return i;}))"
        js    = jsDefault.replace('$docType$', newString).replace('$key$','doc._id')
        jsAll = jsDefault.replace('$docType$', "doc['-type']=='x0'").replace('$key$','doc._id')
      elif docType[0] == 'x':
        continue
      elif docType == '-':
        replaceStrings = []
        for jDocType in normalDocTypes:
          replaceStrings.append( f"doc['-type'].join('/').substring(0,{len(jDocType)})!='{jDocType}'" )
        js = jsDefault.replace('$docType$',
            ' && '.join(replaceStrings) +" && (doc['-branch'][0].show.every(function(i) {return i;}))",
            ).replace('$key$', 'branch.stack[0]')
        jsAll = jsDefault.replace('$docType$',
            ' && '.join(replaceStrings)
            ).replace('$key$', 'branch.stack[0]')
      else: #show all doctypes that have the same starting ..
        js = jsDefault.replace('$docType$',
            f"doc['-type'].join('/').substring(0,{len(docType)})=='{docType}"+"' && (doc['-branch'][0].show.every(function(i) {return i;}))",
            ).replace('$key$', 'branch.stack[0]')
        jsAll = jsDefault.replace('$docType$',
            f"doc['-type'].join('/').substring(0, {len(docType)})=='{docType}'",
            ).replace('$key$', 'branch.stack[0]')
      outputList = []
      baseDocType = docType[:-3] if docType.endswith('All') else docType
      if docTypeChange==docType:
        columnNames = columnsChange
      elif baseDocType in oldColumnNames:
        columnNames = oldColumnNames[baseDocType].split(',')
      elif docType == '-':
        columnNames = [i['name'] for i in defaultDataHierarchyNode['default'] if 'name' in i]
      else:
        columnNames = [i['name'] for group in self.dataHierarchy[docType]['meta']
                       for i in self.dataHierarchy[docType]['meta'][group]]
      columnNames = columnNames[:tableColumnsMax]
      commentString = f'// {docType} : ' + ','.join(columnNames) + '\n'
      for name in columnNames:
        if name == 'image':
          outputList.append('doc.image?doc.image.length>3:false')  #Not as .toString() because that leads to inconsistencies
        elif name == '-tags':
          outputList.append("doc['-tags'].join(' ')")
        elif '#_' in name:
          outputList.append(f'doc["-tags"].indexOf("{name[1:]}")>-1')
        elif name == '-type':
          outputList.append('doc["-type"].slice(1).join("/")')
        elif name == 'content':
          outputList.append('doc.content?doc.content.slice(0, 100):""')
        elif '/' in name:  #stacked requests i.e. metaVendor/date
          parentString = 'doc' + ''.join([f'["{i}"]' for i in name.split('/')[:-1]])
          newString = 'doc' + ''.join([f'["{i}"]' for i in name.split('/')])
          newString = f'{parentString} ? {newString}: ""'
          outputList.append(newString)
        else:
          outputList.append(f'doc["{name}"]')
      outputStr = ','.join(outputList)
      viewCode[docType.replace('/','__')]       = commentString+js.replace('$outputList$', outputStr)
      viewCode[docType.replace('/','__')+'All'] = commentString+jsAll.replace('$outputList$', outputStr)
    self.saveView('viewDocType', viewCode)
    return


  def getColumnNames(self) -> dict[str,str]:
    """ get names of table columns from design documents

    Returns:
      dict: docType and ,-separated list of names as string
    """
    if '_design/viewDocType' not in self.db:
      return {}
    try:
      comments = [v['map'].split('\n// ')[1].split('\n')[0]
                  for v in self.db['_design/viewDocType']['views'].values()]
      return {i.split(' : ')[0]:i.split(' : ')[1] for i in comments}
    except Exception:
      defaultColumnNames = {'-':','.join([i['name'] for i in defaultDataHierarchyNode['default'] if 'name' in i])}
      for docType, docTypeDict in self.dataHierarchy.items():
        if docType[0] in ['_','-']:
          continue
        columnNames = [i['name'] for group in docTypeDict['meta'] for i in docTypeDict['meta'][group]]
        defaultColumnNames[docType] = ','.join(columnNames)
      return defaultColumnNames


  def initGeneralViews(self) -> None:
    """
    general views: Hierarchy, Identify
    """
    tracebackString(True, 'initGeneralView')
    jsHierarchy  = '''
      if ('-type' in doc && (doc["-branch"][0].show.every(function(i) {return i;}))) {
        doc['-branch'].forEach(function(branch, idx) {emit(branch.stack.concat([doc._id]).join(' '),[branch.child,doc['-type'],doc['-name'],doc['-gui'],idx]);});
      }
    '''
    jsHierarchyAll  = '''
      if ('-type' in doc) {
        doc['-branch'].forEach(function(branch, idx) {emit(branch.stack.concat([doc._id]).join(' '),[branch.child,doc['-type'],doc['-name'],doc['-gui'],idx]);});
      }
    '''
    jsPath = '''
      if ('-type' in doc && '-branch' in doc && (doc["-branch"][0].show.every(function(i) {return i;}))){
        if ('shasum' in doc){doc['-branch'].forEach(function(branch,idx){if(branch.path){emit(branch.path,[branch.stack,doc['-type'],branch.child,doc.shasum,idx]);}});}
        else                {doc['-branch'].forEach(function(branch,idx){if(branch.path){emit(branch.path,[branch.stack,doc['-type'],branch.child,''        ,idx]);}});}
      }
    '''
    jsPathAll = '''
      if ('-type' in doc && '-branch' in doc){
        if ('shasum' in doc){doc['-branch'].forEach(function(branch,idx){if(branch.path){emit(branch.path,[branch.stack,doc['-type'],branch.child,doc.shasum,idx]);}});}
        else                {doc['-branch'].forEach(function(branch,idx){if(branch.path){emit(branch.path,[branch.stack,doc['-type'],branch.child,''        ,idx]);}});}
      }
    '''
    self.saveView('viewHierarchy',{'viewHierarchy':jsHierarchy,'viewPaths':jsPath,\
                                   'viewHierarchyAll':jsHierarchyAll,'viewPathsAll':jsPathAll})
    jsSHA= "if (doc['-type'][0]==='measurement'){emit(doc.shasum, doc['-name']);}"
    jsQR = "if (doc.qrCode.length > 0) {doc.qrCode.forEach(function(thisCode) {emit(thisCode, doc['-name']);});}"
    jsTags="if ('-tags' in doc && (doc['-branch'][0].show.every(function(i) {return i;})))"+\
              "{doc['-tags'].forEach(function(tag){emit(tag,[doc['-name'], doc['-type'].join('/')]);});}"
    jsTagsAll="if ('-tags' in doc){doc['-tags'].forEach(function(tag){emit(tag,[doc['-name'], doc['-type'].join('/')]);});}"
    views = {'viewQR':jsQR, 'viewSHAsum':jsSHA, 'viewTags':jsTags, 'viewTagsAll':jsTagsAll}
    self.saveView('viewIdentify', views)
    return



  def exit(self, deleteDB:bool=False) -> None:
    """
    Shutting down things

    Args:
      deleteDB (bool): remove database
    """
    import warnings
    if deleteDB:
      self.db.client.delete_database(self.databaseName)
    warnings.simplefilter("ignore")  #client disconnect triggers ignored ResourceWarning on socket
    self.client.disconnect()
    return


  def getDoc(self, docID:str) -> dict[str,Any]:
    """
    Wrapper for get from database function

    Args:
        docID (dict): document id

    Returns:
        dict: json representation of document
    """
    return dict(self.db[docID])


  def saveDoc(self, doc:dict[str,Any]) -> dict[str,Any]:
    """
    Wrapper for save to database function

    Discussion on -branch['path']:
    - full path (from basePath) allows to easily create a view of all paths and search through them
      during each scan, which happens rather often
    - just the incremental path (file-name, folder-name) allows to easily change that if the user wants
      and not change all the children paths, too. However, the renaming of the folder is likely occurring
      less often.

    Args:
        doc (dict): document to save

    Returns:
        dict: json representation of submitted document
    """
    doc['-client'] = tracebackString(False, 'save:'+doc['_id'])
    if '-branch' in doc and 'op' in doc['-branch']:
      del doc['-branch']['op']  #remove operation, saveDoc creates and therefore always the same
      if 'show' not in doc['-branch']:
        doc['-branch']['show'] = self.createShowFromStack(doc['-branch']['stack'])
      doc['-branch'] = [doc['-branch']]
    if '_attachments' in doc:   # delete if document was moved
      del doc['_attachments']
    try:
      res = self.db.create_document(doc)
      logging.debug('successfully saved doc with type and branch '+doc['_id']+' '+'/'.join(doc['-type'])+'  |  '+str(doc['-branch'])+'\n')
    except Exception:
      logging.error('could not save, likely JSON issue; Check logging file')
      logging.error('could not save, likely JSON issue')
      if 'image' in doc:
        del doc['image']
      logging.error(str(doc))
      logging.error(traceback.format_exc())
      res=None
    return res


  def updateDoc(self, change:dict[str,Any], docID:str) -> dict[str,Any]:
    """
    Update document by
    - saving changes to oldDoc (revision document)
    - updating new-document concurrently
    - create a docID for oldDoc
    - Bonus: save '_rev' from newDoc to oldDoc in order to track that updates cannot happen by accident

    Args:
        change (dict): item to update
        docID (string):  id of document to change

    Returns:
        dict: json representation of updated document
    """
    change['-client'] = tracebackString(False, f'updateDoc:{docID}')
    newDoc = self.db[docID]  #this is the document that stays live
    initialDocCopy = dict(newDoc)
    if 'edit' in change:     #if delete
      oldDoc = dict(newDoc)
      for item in oldDoc:
        if item not in ('_id', '_rev', '-branch'):
          del newDoc[item]
      newDoc['-user']   = change['-user']
    else:                    #if update
      oldDoc = {}            #this is an older revision of the document
      # nothingChanged = True
      # handle branch
      if '-branch' in change:# and len(change['-branch']['stack'])>0:
        op = change['-branch'].pop('op')
        oldpath = change['-branch'].pop('oldpath',None)
        if change['-branch']['path'] is None:
          change['-branch']['path']=newDoc['-branch'][0]['path']
        if change['-branch'] not in newDoc['-branch']:       #skip if new branch is already in branch
          oldDoc['-branch'] = newDoc['-branch'].copy()
          for branch in newDoc['-branch']:
            if op=='c' and branch['path']==change['-branch']['path']:
              op='u'
          if op=='c':    #create, append
            change['-branch']['show'] = self.createShowFromStack(change['-branch']['stack'])
            newDoc['-branch'] += [change['-branch']]
            # nothingChanged = False
          elif op=='u':  #update
            if oldpath is not None:
              for branch in newDoc['-branch']:
                if branch['path'].startswith(oldpath):
                  if os.path.basename(branch['path']) == newDoc['-name'] and \
                         os.path.basename(str(change['-branch']['path']))!='':
                    newDoc['-name'] = os.path.basename(str(change['-branch']['path']))
                  branch['path'] = branch['path'].replace(oldpath ,change['-branch']['path'])
                  branch['stack']= change['-branch']['stack']
                  branch['show'] = self.createShowFromStack(change['-branch']['stack'], branch['show'][-1])
                  break
            else:
              newDoc['-branch'][0] = change['-branch'] #change the initial one
            # nothingChanged = False
          elif op=='d':  #delete
            newDoc['-branch'] = [branch for branch in newDoc['-branch'] if branch['path']!=change['-branch']['path']]
            # originalLength = len(newDoc['-branch'])
            # if originalLength!=len(newDoc['-branch']):
            #   nothingChanged = False
          else:
            logging.warning('database.update.1: unknown branch op: '+newDoc['_id']+' '+newDoc['-name'])
            return newDoc
      #handle other items
      # change has to be dict, not Document
      for item in change:
        if item in ['_id','_rev','-branch']:                #skip items cannot do not result in change
          continue
        if item=='-type' and change['-type']==['--']:          #skip non-set type
          continue
        if item=='image' and change['image']=='':          #skip if non-change in image
          continue
        if change[item] is None:
          # DO CHANGE TO EMPTY: if user wants it
          # or (isinstance(change[item], str) and change[item].strip()=='') or \
          #    (isinstance(change[item], list) and len(change[item])==0):      #skip empty entries
          continue
        ## Discussion: What if content only differs by whitespace changes?
        # These changes should occur in the database, the user wanted it so
        # Do these changes justify a new revision?
        # Hence one could update the doc and previous-revision(with the current _rev)
        #  - but that would lead to special cases, more code, chaos
        #  - also not sure how often simple white space changes occur, how important
        # To identify these cases use the following
        # if (isinstance(change[item], str) and " ".join(change[item].split())!=" ".join(newDoc[item].split()) ) or \
        #    (isinstance(change[item], list) and change[item]!=newDoc[item] ):
        # Add to testBasic to test for it:
        #       myString = myString.replace('A long comment','A long   comment')
        if item not in newDoc or change[item]!=newDoc[item]:
          # if item not in ['-date','-client','-user']:      #if only date/client change, no significant change
          #   nothingChanged = False
          if item == 'image':
            oldDoc[item] = 'image changed'       #don't backup images: makes database big and are only thumbnails anyhow
          elif item in newDoc:
            oldDoc[item] = newDoc[item]
          newDoc[item] = change[item]
      # Always update, for some reason single tags are not recognized
      # if nothingChanged:
      #   logging.debug('database.update.2: doc not updated-nothing changed: '+newDoc['_id']+' '+newDoc['-name'])
      #   return newDoc
    #For both cases: delete and update
    if '_curated' not in newDoc['-tags'] and newDoc['-type'][0][0]!='x':
      newDoc['-tags'].append('_curated')
    try:
      newDoc.save()
    except Exception:
      logging.error('database.update: update unsuccessful: '+newDoc['_id']+' '+newDoc['-name'])
      print('**ERROR: could not update document. Likely version conflict. Initial and current version:')
      print(initialDocCopy)
      print(newDoc)
      return {}
    attachmentName = 'v0.json'
    if '_attachments' in newDoc:
      attachmentName = 'v'+str(len(newDoc['_attachments']))+'.json'
    newDoc.put_attachment(attachmentName, 'application/json', json.dumps(oldDoc))
    logging.debug('database.update.3: doc update success: '+newDoc['_id']+' '+newDoc['-name']+'\n')
    return newDoc


  def updateBranch(self, docID:str, branch:int, child:int, stack:Optional[list[str]]=None,
                   path:Optional[str]='') -> tuple[str, Optional[str]]:
    """
    Update document by updating the branch

    Args:
      docID (string):  id of document to change
      branch (int):  index of branch to change
      child (int):  new number of child
      stack (list):  new list of ids
      path (str): new path; None is acceptable

    Returns:
      str, str: old path, new path
    """
    doc = self.db[docID]
    doc['-client'] = tracebackString(False, f'updateBranch:{docID}')
    if len(doc['-branch'])<=branch:
      print(f'**ERROR Cannot delete branch that does not exist {branch}in doc {docID}')
      logging.error('**ERROR Cannot delete branch that does not exist %s in doc %s', branch, docID)
    oldPath = doc['-branch'][branch]['path']
    if oldPath is None:
      path = None
    if path=='':
      name = f'{child:03d}_' + '_'.join(oldPath.split('/')[-1].split('_')[1:])
      path = '/'.join(oldPath.split('/')[:-1]+[name])
    # test if path already exists
    if docID[0]=='x' and path is not None:
      if not (self.basePath/path/'.id_pastaELN.json').is_file():
        logging.debug('Target folder\'s json does not exist: '+path)
        oldDocID = ''
      else:
        with open(self.basePath/path/'.id_pastaELN.json', 'r', encoding='utf-8') as fIn:
          oldJsonContent = json.load(fIn)
          oldDocID = oldJsonContent['_id']
      if path is not None and (self.basePath/path).exists() and docID!=oldDocID:
        logging.info('Target folder already exist: %s. I will try to create a new path name', path)
      while path is not None and (self.basePath/path).exists() and docID!=oldDocID:
        if re.search(r"_\d+$", path) is None:
          path += '_1'
        else:
          path = '_'.join(path.split('_')[:-1])+'_'+str(int(path.split('_')[-1])+1)
    # assemble data
    doc['-branch'][branch]['path']=path
    doc['-branch'][branch]['child']=child
    stackOld = list(doc['-branch'][branch]['stack'])
    if stack is not None:
      doc['-branch'][branch]['stack']=stack
    doc['-branch'][branch]['show'] = self.createShowFromStack(doc['-branch'][branch]['stack'],
                                                              doc['-branch'][branch]['show'][-1])
    doc.save()
    logging.debug('success BRANCH updated with type and branch '+doc['_id']+' '+'/'.join(doc['-type'])+'  |  '+str(doc['-branch'])+'\n')
    # move content: folder and data and write .json to disk
    if oldPath is not None and path is not None and ':/' not in oldPath:
      if not (self.basePath/oldPath).exists() and (self.basePath/path).exists():
        logging.debug('database:updateBranch: dont move since already good')
      else:
        (self.basePath/oldPath).rename(self.basePath/path)
    if docID[0]=='x' and path is not None:
      with open(self.basePath/path/'.id_pastaELN.json', 'w', encoding='utf-8') as fOut:
        fOut.write(json.dumps(self.getDoc(docID)))
      # update children's paths
      children = self.getView('viewHierarchy/viewHierarchy', startKey=' '.join(stackOld+[docID,'']))
      for line in children:
        docLine = self.db[line['id']]
        for branchLine in docLine['-branch']:
          if branchLine['path'] is not None and branchLine['path'].startswith(oldPath):
            branchLine['path'] = path+branchLine['path'][len(oldPath):]
          if stack is not None and '/'.join(branchLine['stack']).startswith('/'.join(stackOld)):
            branchLine['stack'] = stack+branchLine['stack'][len(stackOld):]
            branchLine['show']  = self.createShowFromStack(branchLine['stack'], branchLine['show'][-1])
        docLine.save()
        # update .json on disk
        for branchLine in docLine['-branch']:
          if line['id'][0]=='x'  and (self.basePath/branchLine['path']).is_dir():
            with open(self.basePath/branchLine['path']/'.id_pastaELN.json', 'w', encoding='utf-8') as fOut:
              fOut.write(json.dumps(docLine))
    return oldPath, path


  def createShowFromStack(self, stack:list[str], currentShow:bool=True) -> list[bool]:
    """
    For branches: create show entry in the branches by using the stack
    - should be 1 longer than stack
    - check parents if hidden, then this child is hidden too

    Args:
      stack (list): list of ancestor docIDs
      currentShow (bool): current show-indicator of this item

    Returns:
      list: list of show = list of bool
    """
    show = len(stack)*[True] + [currentShow]
    for idx, docID in enumerate(stack):
      if not self.db[docID]['-branch'][0]['show'][-1]:
        show[idx] = False
    return show


  def remove(self, docID:str) -> dict[str,Any]:
    """
    remove doc from database: temporary for development and testing

    Args:
      docID (string): id of document to remove

    Returns:
      dict: document that was removed
    """
    tracebackString(True, f'remove:{docID}')
    doc = self.db[docID]
    res = dict(doc)
    doc.delete()
    return res


  def addAttachment(self, docID:str, name:str, content:dict[str,Any]) -> bool:
    """
    Update document by adding attachment (no new revision)

    Args:
        docID (string):  id of document to change
        name (string): attachment name to add to
        content (dict): dictionary of content to be added (should include user,date,docID,remark)

    Returns:
        bool: success of method
    """
    try:
      doc = self.db[docID]
      if '-attachment' not in doc:
        doc['-attachment'] = {}
      if name in doc['-attachment']:
        doc['-attachment'][name] += [content]
      else:
        doc['-attachment'][name] = [content]
      doc.save()
      return True
    except Exception:
      return False


  def getView(self, thePath:str, startKey:Optional[str]=None, preciseKey:Optional[str]=None) -> list[dict[str,Any]]:
    """
    Wrapper for getting view function

    Args:
        thePath (string): path to view
        startKey (string): if given, use to filter output, everything that starts with this key
        preciseKey (string): if given, use to filter output. Match precisely

    Returns:
        list: list of documents in this view
    """
    from cloudant.view import View
    designDoc = self.db.get_design_document(thePath.split('/')[0])
    v = View(designDoc, thePath.split('/')[1])
    try:
      if startKey is not None:
        res = v(startkey=startKey, endkey=f'{startKey}zzz')['rows']
      elif preciseKey is not None:
        res = v(key=preciseKey)['rows']
      else:
        res = list(v.result)
    except Exception:
      print('**ERROR dgv01: Database / Network problem for path |', thePath)
      print(traceback.format_exc())
      res = []
    return res


  def saveView(self, designName:str, viewCode:dict[str,str]) -> None:
    """
    Adopt the view by defining a new jsCode

    Args:
        designName (string): name of the design
        viewCode (dict): viewName: js-code
    """
    from cloudant.design_document import DesignDocument
    if f'_design/{designName}' in self.db:
      designDoc = self.db[f'_design/{designName}']
      if '_rev' in designDoc:
        designDoc.delete()
    designDoc = DesignDocument(self.db, designName)
    for view in viewCode:
      thisJsCode = 'function (doc) {\n' + viewCode[view] + '\n}'
      designDoc.add_view(view, thisJsCode)
    try:
      designDoc.save()
    except Exception:
      print('**ERROR dsv01: something unexpected has happened.')
      logging.error('dsv01: something unexpected has happened.')
      logging.error(traceback.format_exc())
    return


  def getHierarchy(self, start:str, allItems:bool=False) -> Node:
    """
    get hierarchy tree for projects, ...

    Args:
      start (str): start of the hierarchy (most parent)
      allItems (bool):  true=show all items, false=only non-hidden

    Returns:
      Node: hierarchy in an anytree
    """
    if not allItems:
      view = self.getView('viewHierarchy/viewHierarchy',    startKey=start)
    if allItems or len(view)==0:
      view = self.getView('viewHierarchy/viewHierarchyAll', startKey=start)
    # for item in view:
    #   print(item)
    # Reorganize data into lists
    childNum = {i['id']:i['value'][0] for i in view}
    # ids = [i['id'] for i in view]
    keys = [i['key'] for i in view]
    values = [i['value'] for i in view]
    for k,v in childNum.items():
      keys = [i.replace(k,f'{v} {k}') for i in keys]
    values = [x for _, x in sorted(zip(keys, values))]
    keys   = sorted(keys)
    dataTree = None
    hierarchy = []
    for idx, value in enumerate(values):
      docType = value[1]
      name    = value[2]
      gui     = value[3]
      _id     = keys[idx].split()[-1]
      level   = int(len(keys[idx].split())/2)
      if idx==0:
        dataTree = Node(id=_id, docType=docType, name=name, gui=gui)
        hierarchy.append(dataTree)
      else:
        parentNode = hierarchy[level-2]
        subNode = Node(id=_id, parent=parentNode, docType=docType, name=name, gui=gui)
        if len(hierarchy)<level:
          hierarchy.append('')
        elif len(hierarchy)>level:
          hierarchy.pop()
        hierarchy[-1] = subNode
    return dataTree


  def hideShow(self, stack:Union[str,list[str]]) -> None:
    """
    Toggle hide/show indicator of branch

    Args:
      stack (list, str): stack of docID; docID (str)
    """
    flippedOnce = False
    if isinstance(stack, str):
      doc = self.db[stack]
      for idx, _ in enumerate(doc['-branch']):
        doc['-branch'][idx]['show'][-1] = not doc['-branch'][idx]['show'][-1]
        logging.debug('flipped str: %s',str(stack))
      doc.save()
      if stack[0]=='x':
        stack = doc['-branch'][0]['stack']+[stack]
      flippedOnce = True
    if isinstance(stack, list):
      iFlip = len(stack)-1
      logging.debug('  database list %s %s',str(stack),str(iFlip))
      for item in self.getView('viewHierarchy/viewHierarchyAll', startKey=' '.join(stack)):
        logging.debug('  docID: %s',item['id'])
        doc = self.db[item['id']]
        for idx, branch in enumerate(doc['-branch']):
          if not flippedOnce or iFlip!=len(branch['stack']):
            doc['-branch'][idx]['show'][iFlip] = not doc['-branch'][idx]['show'][iFlip]
        doc.save()
    if doc['-type'][0][0]=='x':
      with open(self.basePath/doc['-branch'][0]['path']/'.id_pastaELN.json','w', encoding='utf-8') as fOut:
        fOut.write(json.dumps(doc))
    return


  def setGUI(self, docID:str, guiState:list[bool]) -> None:
    """
    Set the gui state
    - 0: true=show details; false=hide details
    - 1: true=show children; false=hide children; only makes sense for folders: doctype = x1

    Args:
      docID (str): docID
      guiState (list): list of bool that show if document is shown
    """
    doc = self.db[docID]
    doc['-gui'] = guiState
    doc.save()
    return


  def replicateDB(self, dbInfo:dict[str,Any], progressBar:Any, removeAtStart:bool=False) -> str:
    """
    Replication to another instance

    One cannot simply create the document no the other server...
    - because then the _rev do not match and will never sync
    - issues arise if the documents in a database are deleted. Better remove the entire database and recreate it

    Args:
        dbInfo (dict): info on the remote database
        progressBar (QProgressBar): gui - qt progress bar
        removeAtStart (bool): remove remote DB before starting new

    Returns:
        str: report
    """
    progressBar.show()
    try:
      rep = Replicator(self.client)
      try:
        client2 = CouchDB(dbInfo['user'], dbInfo['password'], url=dbInfo['url'], connect=True)
      except Exception:
        return '<b>ERROR drp01: Could not connect to remote server. Abort replication.</b><br>'+\
                   'user:'+dbInfo['user']+'<br>password:'+dbInfo['password']+'<br>url:'+dbInfo['url']
      db2 = client2[dbInfo['database']]
      replResult = rep.create_replication(self.db, db2, create_target=False, continuous=False)
      logging.info('Start replication '+replResult['_id']+'.')
      progressBar.setValue(10)
      #try every 10sec whether replication success. Do that for max. of 5min
      startTime = time.time()
      while True:
        progressBar.setValue(10+ int((time.time()-startTime)/60./5.*90) )
        if (time.time()-startTime)/60.>5.:
          logging.info('Stop waiting for replication '+replResult['_id']+'.')
          progressBar.hide()
          return "Waited for 5min. No replication success in that time"
        replResult.fetch()        # get updated, latest version from the server
        if '_replication_state' in replResult:
          logging.info('Success replication '+replResult['_id']+'.')
          progressBar.hide()
          reply = (f"Replication success state: {replResult['_replication_state']}\n  "
                   f"time of reporting: {replResult['_replication_state_time']}\n"
                   f"Documents in source database: {len(list(self.db))}\n")
          stats = dict(replResult['_replication_stats'])
          del stats['checkpointed_source_seq']
          reply += '\n  '.join([f"{k.replace('_',' ')}:{v}" for k,v in stats.items()])+'\n'
          # check replication successful
          setLocal = {i['_id'] for i in self.db}
          setRemote= {i['_id'] for i in db2}
          if setLocal.difference(setRemote):
            reply += f' Difference local-global {setLocal.difference(setRemote)}\n'
          if setRemote.difference(setLocal):
            reply += f' Difference global-local {setRemote.difference(setLocal)}\n'
          return reply
        time.sleep(10)
    except Exception:
      progressBar.hide()
      return "**ERROR drp02: replicate error |\n"+traceback.format_exc()



  def historyDB(self) -> dict[str,Any]:
    """
    Collect last modification days of documents
    """
    from datetime import datetime
    import numpy as np
    collection:dict[str,Any] = {}
    for doc in self.db:
      if doc['_id'][1]=='-' and len(doc['_id'])==34 and '-type' in doc and '-date' in doc:
        docType = doc['-type'][0]
        date = doc['-date'][:-1]
        if len(date)==22:
          date += '0'
        date    = datetime.fromisoformat( date ).timestamp()
        if docType in collection:
          collection[docType] = collection[docType] + [date]
        else:
          collection[docType] = [date]
    #determine bins for histogram
    firstSubmit = datetime.now().timestamp()
    for value in collection.values():
      firstSubmit = min(firstSubmit, np.min(value))
    bins = np.linspace(firstSubmit, datetime.now().timestamp(), 100 )
    #calculate histogram and save it
    collectionCopy = dict(collection)
    for key,value in collection.items():
      hist, _ = np.histogram(value, bins)
      collectionCopy[key] = hist
    collectionCopy['-bins-'] = (bins[:-1]+bins[1:])/2
    #calculate score
    bias = np.exp(( collectionCopy['-bins-']-collectionCopy['-bins-'][-1] ) / 1.e7)
    score = {key: np.sum(value*bias) for key, value in collectionCopy.items()}
    #reformat dates into string
    collectionCopy['-bins-'] = [datetime.fromtimestamp(i).isoformat() for i in collectionCopy['-bins-']]
    collectionCopy['-score-']= score
    return collectionCopy


  def checkDB(self, outputStyle:str='text', repair:bool=False, minimal:bool=False) -> str:
    """
    Check database for consistencies by iterating through all documents
    - slow since no views used
    - check views
    - only reporting, no repair
    - custom changes are possible with normal scan
    - no interaction with harddisk

    Args:
        outputStyle (str): output using a given style: see outputString
        repair (bool): repair database
        minimal (bool): true=only show warnings and errors; else=also show information

    Returns:
        str: output
    """
    outstring = ''
    if outputStyle=='html':
      outstring += '<div align="right">'
    outstring+= outputString(outputStyle,'h2','LEGEND')
    if not minimal:
      outstring+= outputString(outputStyle,'ok','Green: perfect and as intended')
      outstring+= outputString(outputStyle,'okish', 'Blue: ok-ish, can happen: empty files for testing, strange path for measurements')
    outstring+= outputString(outputStyle,'unsure', 'Dark magenta: unsure if bug or desired (e.g. move step to random path-name)')
    outstring+= outputString(outputStyle,'warning','Orange-red: WARNING should not happen (e.g. procedures without project)')
    outstring+= outputString(outputStyle,'error',  'Red: FAILURE and ERROR: NOT ALLOWED AT ANY TIME')
    if outputStyle=='html':
      outstring += '</div>'
    outstring+= outputString(outputStyle,'h2','List all database entries')
    if repair:
      print('REPAIR MODE IS ON: afterwards, full-reload and create views')
    for doc in self.db:
      try:
        if '_design' in doc['_id']:
          if not minimal:
            outstring+= outputString(outputStyle,'ok','..info: Design document '+doc['_id'])
          continue
        if doc['_id'] == '-dataHierarchy-':
          if repair:
            if '-hierarchy-' in doc:
              del doc['-hierarchy-']
            for old,new in [['project','x0'],['step','x1'],['task','x2']]:
              if new not in doc and old in doc:
                doc[new] = doc[old].copy()
                del doc[old]
            doc.save()
          if not minimal:
            outstring+= outputString(outputStyle,'ok','..info: dataHierarchy exists')
          continue
        #only normal documents after this line

        ###custom temporary changes: keep few as examples;
        # BE CAREFUL: PRINT FIRST, delete second run ; RUN ONLY ONCE
        #   Project renaming bug 1)
        # if doc['-branch'][0]['path'].startswith('PastaEln'):
        #   oldPath = doc['-branch'][0]['path']
        #   newPath = oldPath.replace('PastaEln', 'Pasta')
        #   print(oldPath, '->', newPath)
        #   doc['-branch'][0]['path'] = newPath
        #   doc.save()

        #branch test
        if '-branch' not in doc:
          outstring+= outputString(outputStyle,'error',f"dch01: branch does not exist {doc['_id']}")
          continue
        if len(doc['-branch'])>1 and doc['-type'] =='x':                 #text elements only one branch
          outstring+= outputString(outputStyle,'error',f"dch02: branch length >1 for text {doc['_id']} {str(doc['-type'])}")
        for branch in doc['-branch']:
          for item in branch['stack']:
            if not item.startswith('x-'):
              outstring+= outputString(outputStyle,'error',f"dch03: non-text in stack {doc['_id']}")

          if len(branch['stack'])==0 and doc['-type']!=['x0']: #if no inheritance
            if doc['-type'][0] == 'measurement' or  doc['-type'][0][0] == 'x':
              outstring+= outputString(outputStyle,'warning',f"branch stack length = 0: no parent {doc['_id']}")
            elif not minimal:
              outstring+= outputString(outputStyle,'okish',f"branch stack length = 0: no parent for procedure/sample {doc['_id']} | {doc['-name']}")
          if '-type' not in doc or len(doc['-type'])==0:
            outstring+= outputString(outputStyle,'unsure',f"dch04: no type in (removed data?) {doc['_id']}")
            continue
          if doc['-type'][0][0]=='x':
            try:
              dirNamePrefix = branch['path'].split(os.sep)[-1].split('_')[0]
              if dirNamePrefix.isdigit() and branch['child']!=int(dirNamePrefix): #compare child-number to start of directory name
                outstring+= outputString(outputStyle,'error',f"dch05: child-number and dirName dont match {doc['_id']}")
            except Exception:
              pass  #handled next lines
          if branch['path'] is None:
            if doc['-type'][0][0] == 'x':
              outstring+= outputString(outputStyle,'error',f"dch06: branch path is None {doc['_id']}")
            elif doc['-type'][0] == 'measurement':
              if not minimal:
                outstring+= outputString(outputStyle,'okish','measurement branch path is None=no data '+doc['_id']+' '+doc['-name'])
            elif not minimal:
              outstring+= outputString(outputStyle,'ok',f"procedure/sample with empty path {doc['_id']}")
          else:                                                    #if sensible path
            if len(branch['stack'])+1 != len(branch['path'].split(os.sep)):#check if length of path and stack coincide
              if doc['-type'][0] == 'procedure':
                if not minimal:
                  outstring+= outputString(outputStyle,'ok',f"procedure: branch stack and path lengths not equal: {doc['_id']} | {branch['path'][:30]}")
              else:
                outstring+= outputString(outputStyle,'unsure',f"branch stack and path lengths not equal: {doc['_id']} | {branch['path'][:30]}")
            if branch['child'] != 9999:
              for parentID in branch['stack']:                  #check if all parents in doc have a corresponding path
                parentDoc = self.getDoc(parentID)
                if '-branch' not in parentDoc:
                  outstring += outputString(outputStyle, 'error', f'dch07: branch not in parent with id {parentID}')
                  continue
                parentDocBranches = parentDoc['-branch']
                onePathFound = any(parentBranch['path'] is not None and parentBranch['path'] in branch['path'] for parentBranch in parentDocBranches)
                if not onePathFound:
                  outstring+= outputString(outputStyle,'unsure',f"dch08: parent does not have corresponding path (remote content) {doc['_id']} | parentID {parentID}")
          if 'show' not in branch:
            outstring+= outputString(outputStyle,'error',f"dch08a: branch does not have show: {doc['_id']}")
          elif len(branch['show']) != len(branch['stack'])+1:
            outstring+= outputString(outputStyle,'error',f"dch08b: branch-show not same length as branch-stack: {doc['_id']}")

        #every doc should have these:
        for key in ['-name','-tags','-client','-user','-date']:
          if key not in doc:
            outstring+= outputString(outputStyle,'error',f"dch17: {key} not in (deleted doc?){doc['_id']}")

        #doc-type specific tests
        if '-type' in doc and doc['-type'][0] == 'sample':
          if 'qrCode' not in doc:
            outstring+= outputString(outputStyle,'warning',f"dch09: qrCode not in sample {doc['_id']}")
        elif '-type' in doc and doc['-type'][0] == 'measurement':
          if 'shasum' not in doc:
            outstring+= outputString(outputStyle,'warning',f"dch10: shasum not in measurement {doc['_id']}")
          if 'image' not in doc:
            outstring+= outputString(outputStyle,'warning',f"dch11: image not in measurement {doc['_id']}")
          else:
            if doc['image'].startswith('data:image'):  #for jpg and png
              try:
                imgdata = base64.b64decode(doc['image'][22:])
                Image.open(io.BytesIO(imgdata))  #can convert, that is all that needs to be tested
              except Exception:
                outstring+= outputString(outputStyle,'error',f"dch12: jpg-image not valid {doc['_id']}")
            elif doc['image'].startswith('<?xml'):
              #from https://stackoverflow.com/questions/63419010/check-if-an-image-file-is-a-valid-svg-file-in-python
              SVG_R = r'(?:<\?xml\b[^>]*>[^<]*)?(?:<!--.*?-->[^<]*)*(?:<svg|<!DOCTYPE svg)\b'
              SVG_RE = re.compile(SVG_R, re.DOTALL)
              if SVG_RE.match(doc['image']) is None:
                outstring+= outputString(outputStyle,'error',f"dch13: svg-image not valid {doc['_id']}")
            elif doc['image']=='':
              outstring+= outputString(outputStyle,'unsure',f"image not valid {doc['_id']} {doc['image']}")
            else:
              outstring+= outputString(outputStyle,'error',f"dch14: image not valid {doc['_id']} {doc['image']}")

      except Exception:
        outstring+= outputString(outputStyle,'error', f"dch15: critical error in {doc['_id']}\n {traceback.format_exc()}")

    ##TEST views
    outstring+= outputString(outputStyle,'h2','List problematic database tables')
    view = self.getView('viewIdentify/viewSHAsum')
    shasumKeys = []
    for item in view:
      if item['key']=='':
        outstring+= outputString(outputStyle,'warning', f"measurement without shasum: {item['id']} {item['value']}")
      else:
        if item['key'] in shasumKeys:
          key = item['key'] or '- empty string -'
          outstring += outputString(outputStyle, 'error', f"dch16: shasum twice in view: {key} {item['id']} {item['value']}")
        shasumKeys.append(item['key'])
    return outstring



#!/usr/bin/python3
"""TEST using the FULL set of python-requirements: create 3 projects; simplified form of testTutorialComplex """
import os, shutil, traceback, logging
from datetime import datetime
import warnings
import unittest
from pathlib import Path
import pytest
from pasta_eln.backend import Backend
from pasta_eln.miscTools import outputString
from pasta_eln.miscTools import DummyProgressBar

class TestStringMethods(unittest.TestCase):
  """
  derived class for this test
  """
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.be = None
    self.dirName = ''

  @pytest.mark.skip(
    reason="Disabled for github since cannot create couchdb instance during actions")
  def test_main(self):
    """
    main function
    """
    dummyProgressBar = DummyProgressBar()
    outputFormat = 'print'
    # initialization: create database, destroy on filesystem and database and then create new one
    warnings.filterwarnings('ignore', message='numpy.ufunc size changed')
    warnings.filterwarnings('ignore', message='invalid escape sequence')
    warnings.filterwarnings('ignore', category=ResourceWarning, module='PIL')
    warnings.filterwarnings('ignore', category=ImportWarning)
    logPath = Path.home()/'pastaELN.log'
    logging.basicConfig(filename=logPath, level=logging.INFO, format='%(asctime)s|%(levelname)s:%(message)s',
                        datefmt='%m-%d %H:%M:%S')   #This logging is always info, since for installation only
    for package in ['urllib3', 'requests', 'asyncio', 'PIL', 'matplotlib.font_manager']:
      logging.getLogger(package).setLevel(logging.WARNING)
    logging.info('Start 3Projects test')

    projectGroup = 'research'
    self.be = Backend('research', initConfig=False)
    self.dirName = self.be.basePath
    self.be.exit(deleteDB=True)
    shutil.rmtree(self.dirName)
    os.makedirs(self.dirName)
    self.be = Backend(projectGroup, initViews=True, initConfig=False)

    try:
      ### CREATE PROJECTS AND SHOW
      outputString(outputFormat,'h2','CREATE PROJECTS AND SHOW')
      self.be.addData('x0', {'-name': 'Intermetals at interfaces', \
        'objective': 'Does spray coating lead to intermetalic phase?', 'status': 'active', \
        'comment': '#intermetal #Fe #Al This is a test project'})
      self.be.addData('x0', {'-name': 'Surface evolution in tribology', \
        'objective': 'Why does the surface get rough during tribology?', 'status': 'passive', \
        'comment': '#tribology The answer is obvious'})
      self.be.addData('x0', {'-name': 'Steel', 'objective': 'Test strength of steel', 'status': 'paused', \
        'comment': '#Fe Obvious example without answer'})
      outputString(outputFormat, 'info', self.be.output('x0'))

      ### TEST PROJECT PLANING
      outputString(outputFormat,'h2','TEST PROJECT PLANING')
      viewProj = self.be.db.getView('viewDocType/x0')
      projID1  = [i['id'] for i in viewProj if i['value'][0]=='Intermetals at interfaces'][0]
      projID2  = [i['id'] for i in viewProj if i['value'][0]=='Steel'][0]

      self.be.changeHierarchy(projID2)
      self.be.addData('x1',    {'comment': 'Random text', '-name': 'Task 1'})
      self.be.addData('x1',    {'-name': 'Task2'})
      self.be.changeHierarchy(None)

      self.be.changeHierarchy(projID1)
      self.be.addData('x1',    {'comment': 'This is hard! #TODO ', '-name': 'Get steel and Al-powder'})
      currentID = self.be.addData('x1',    {'comment': 'This will take a long time. #WAIT #_curated', '-name': 'Get spray machine'})
      self.be.changeHierarchy(currentID)
      self.be.addData('x2',    {'-name': 'Get quotes', 'comment': 'Dont forget company-A #_1 ', 'procedure': 'Guidelines of procurement'})
      self.be.addData('x2',    {'-name': 'Buy machine','comment': 'Delivery time will be 6month #_3 '})
      self.be.changeHierarchy(None)
      semStepID = self.be.addData('x1',    {'-name': 'SEM images'})
      self.be.changeHierarchy(semStepID)
      semDirName = self.be.basePath/self.be.cwd
      self.be.changeHierarchy(None)
      currentID = self.be.addData('x1',    {'-name': 'Nanoindentation'})
      self.be.changeHierarchy(currentID)
      indentDirName = self.be.basePath/self.be.cwd
      self.be.changeHierarchy(None)
      outputString(outputFormat,'info',self.be.outputHierarchy())

      ### TEST PROCEDURES
      outputString(outputFormat,'h2','TEST PROCEDURES')
      sopDir = self.dirName/'StandardOperatingProcedures'
      os.makedirs(sopDir)
      with open(sopDir/'Nanoindentation.md','w', encoding='utf-8') as fOut:
        fOut.write('# Put sample in nanoindenter\n# Do indentation\nDo not forget to\n- calibrate tip\n- *calibrate stiffness*\n')
      with open(sopDir/'SEM.md','w', encoding='utf-8') as fOut:
        fOut.write('# Put sample in SEM\n# Do scanning\nDo not forget to\n- contact the pole-piece\n- **USE GLOVES**\n')
      self.be.addData('procedure', {'-name': 'StandardOperatingProcedures/SEM.md', 'comment': '#v1'})
      self.be.addData('procedure', {'-name': 'StandardOperatingProcedures/Nanoindentation.md', 'comment': '#v1'})
      outputString(outputFormat,'info',self.be.output('procedure'))

      ### TEST SAMPLES
      outputString(outputFormat,'h2','TEST SAMPLES')
      self.be.addData('sample',    {'-name': 'AlFe cross-section', 'chemistry': 'Al99.9; FeMnCr ', 'qrCode': '13214124 99698708', 'comment': 'after OPS polishing'})
      outputString(outputFormat,'info',self.be.output('sample'))
      outputString(outputFormat,'info',self.be.outputQR())

      ###  TEST MEASUREMENTS AND SCANNING
      outputString(outputFormat,'h2','TEST MEASUREMENTS AND SCANNING')
      examplePath = Path(__file__).parent.parent/'pasta_eln'/'Resources'/'ExampleMeasurements'
      shutil.copy(examplePath/'simple.png', semDirName)
      shutil.copy(examplePath/'simple.csv', semDirName)
      shutil.copy(examplePath/'story.odt',  semDirName)
      progressBar = DummyProgressBar()
      self.be.scanProject(progressBar, projID1)

      ### USE GLOBAL FILES
      outputString(outputFormat,'h2','USE GLOBAL FILES')
      self.be.changeHierarchy(semStepID)
      self.be.addData('measurement', {'-name': 'https://upload.wikimedia.org/wikipedia/commons/a/a4/Misc_pollen.jpg'})
      outputString(outputFormat,'info',self.be.output('measurement'))

      ### ADD INSTRUMENTS AND THEIR ATTACHMENTS
      outputString(outputFormat,'h2','ADD INSTRUMENTS AND ATTACHMENTS')
      self.be.addData('instrument', {'-name': 'G200X', 'vendor':'KLA', 'model':'KLA G200X'})
      self.be.addData('instrument', {'-name': 'B1', 'vendor':'Synthon', 'model':'Berkovich tip'})
      output = self.be.output('instrument',True)
      idKLA, idSynthon = None, None
      for line in output.split('\n'):
        if 'KLA' in line:
          idKLA = line.split('|')[-1].strip()
        if 'Synthon'  in line:
          idSynthon = line.split('|')[-1].strip()
      self.be.db.addAttachment(idKLA, "Right side of instrument",
        {'date':datetime.now().isoformat(),'remark':'Worked well','docID':idSynthon,'user':'nobody'})
      self.be.db.addAttachment(idKLA, "Right side of instrument",
        {'date':datetime.now().isoformat(),'remark':'Service','docID':'','user':'nobody'})
      outputString(outputFormat,'info',self.be.output('instrument'))


      ### VERIFY DATABASE INTEGRITY
      outputString(outputFormat,'h2','VERIFY DATABASE INTEGRITY')
      outputString(outputFormat,'info', self.be.checkDB(outputStyle='text'))
      outputString(outputFormat,'h2','DONE WITH VERIFY')
    except:
      outputString(outputFormat,'h2','ERROR OCCURRED IN VERIFY TESTING\n'+ traceback.format_exc() )
      raise
    return

  def tearDown(self):
    logging.info('End 3Projects test')
    return

if __name__ == '__main__':
  unittest.main()



""" Exception raised for configuration errors (reading/writing PASTA config file). """


#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: config_error.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.

class ConfigError(Exception):
  """
  Exception raised for configuration errors (reading/writing PASTA config file).

  Explanation:
  This class represents a custom exception for configuration errors. It is raised when there is an error in the configuration.

  Args:
      message (str): The error message.
      detailed_errors (dict[str, str] | None): Optional detailed error information.

  Attributes:
      message (str): The error message.
      detailed_errors (dict[str, str] | None): Optional detailed error information.
  """

  def __init__(self,
               message: str,
               detailed_errors: dict[str, str] | None = None) -> None:
    """
    Initializes a ConfigError exception.

    Explanation:
    This class represents a custom exception for configuration errors. It is raised when there is an error in the configuration.

    Args:
        message (str): The error message.
        detailed_errors (dict[str, str] | None): Optional detailed error information.

    Attributes:
        message (str): The error message.
        detailed_errors (dict[str, str] | None): Optional detailed error information.
    """
    super().__init__(message)
    self.message = message
    self.detailed_errors = detailed_errors



""" Represents the main dialog for dataverse integration. """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2023
#
#  Author: Jithu Murugan
#  Filename: main_dialog.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.
import datetime
import logging
import textwrap
import time
from typing import Any

import qtawesome as qta
from PySide6 import QtWidgets
from PySide6.QtWidgets import QFrame, QWidget

from pasta_eln.GUI.dataverse.completed_uploads import CompletedUploads
from pasta_eln.GUI.dataverse.dialog_extension import DialogExtension
from pasta_eln.GUI.dataverse.edit_metadata_dialog import EditMetadataDialog
from pasta_eln.GUI.dataverse.main_dialog_base import Ui_MainDialogBase
from pasta_eln.GUI.dataverse.project_item_frame_base import Ui_ProjectItemFrame
from pasta_eln.GUI.dataverse.upload_config_dialog import UploadConfigDialog
from pasta_eln.GUI.dataverse.upload_widget_base import Ui_UploadWidgetFrame
from pasta_eln.dataverse.data_upload_task import DataUploadTask
from pasta_eln.dataverse.database_api import DatabaseAPI
from pasta_eln.dataverse.project_model import ProjectModel
from pasta_eln.dataverse.task_thread_extension import TaskThreadExtension
from pasta_eln.dataverse.upload_model import UploadModel
from pasta_eln.dataverse.upload_queue_manager import UploadQueueManager


class MainDialog(Ui_MainDialogBase):
  """
  Represents the main dialog.

  Explanation:
      This class represents the main dialog of the application.
      It provides methods to handle UI events and manage the upload process.

  """

  def __new__(cls, *_: Any, **__: Any) -> Any:
    """
    Creates a new instance of the MainDialog class.

    Explanation:
        This method creates a new instance of the MainDialog class.

    Args:
        *_: Variable length argument list.
        **__: Arbitrary keyword arguments.

    Returns:
        Any: The new instance of the MainDialog class.
    """
    return super(MainDialog, cls).__new__(cls)

  def __init__(self) -> None:
    """
    Initializes the MainDialog.

    Explanation:
        This method initializes the MainDialog class and sets up the UI.
        It also initializes the database and populates the project widgets.

    """
    self.logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")
    self.instance = DialogExtension()
    super().setupUi(self.instance)
    self.db_api = DatabaseAPI()
    for project in self.db_api.get_models(ProjectModel):
      if isinstance(project, ProjectModel):
        widget = self.get_project_widget(project)
        self.projectsScrollAreaVerticalLayout.addWidget(widget)
    self.uploadPushButton.clicked.connect(self.start_upload)
    self.clearFinishedPushButton.clicked.connect(self.clear_finished)
    self.selectAllPushButton.clicked.connect(lambda: self.select_deselect_all_projects(True))
    self.deselectAllPushButton.clicked.connect(lambda: self.select_deselect_all_projects(False))
    self.config_upload_dialog = UploadConfigDialog()
    self.completed_uploads_dialog = CompletedUploads()
    self.edit_metadata_dialog = EditMetadataDialog()
    self.configureUploadPushButton.clicked.connect(self.show_configure_upload)
    self.showCompletedPushButton.clicked.connect(self.show_completed_uploads)
    self.editFullMetadataPushButton.clicked.connect(self.show_edit_metadata)
    self.upload_manager_task = UploadQueueManager()
    self.config_upload_dialog.config_reloaded.connect(self.upload_manager_task.set_concurrent_uploads)
    self.upload_manager_task_thread = TaskThreadExtension(self.upload_manager_task)
    self.cancelAllPushButton.clicked.connect(self.upload_manager_task.cancel.emit)
    self.buttonBox.button(QtWidgets.QDialogButtonBox.Cancel).clicked.connect(self.close_ui)
    self.instance.closed.connect(self.close_ui)

  def get_upload_widget(self, project_name: str = "") -> dict[str, QFrame | Ui_UploadWidgetFrame]:
    """
    Retrieves the upload widget for a project.

    Explanation:
        This method retrieves the upload widget for a specific project.
        It creates a QFrame instance and sets up the UI for the upload widget.

    Args:
        project_name (str): The name of the project.

    Returns:
        dict[str, QFrame | Ui_UploadWidgetFrame]: The upload widget.
    """
    upload_widget_frame = QtWidgets.QFrame()
    upload_widget_ui = Ui_UploadWidgetFrame()
    upload_widget_ui.setupUi(upload_widget_frame)
    upload_widget_ui.uploadProjectLabel.setText(textwrap.fill(project_name, 45, max_lines=1))
    upload_widget_ui.uploadProjectLabel.setToolTip(project_name)
    upload_widget_ui.statusIconLabel.setPixmap(
      qta.icon('ph.queue-light').pixmap(upload_widget_ui.statusIconLabel.size()))
    upload_widget_ui.logConsoleTextEdit.hide()
    upload_widget_ui.showLogPushButton.clicked.connect(lambda: self.show_hide_log(upload_widget_ui.showLogPushButton))
    upload_widget_ui.modelIdLabel.hide()
    return {"base": upload_widget_frame, "widget": upload_widget_ui}

  def get_project_widget(self, project: ProjectModel) -> QWidget:
    """
    Retrieves the project widget.

    Explanation:
        This method retrieves the project widget for a specific project.
        It creates a QWidget instance and sets up the UI for the project widget.

    Args:
        project (ProjectModel): The project model representing the project.

    Returns:
        QWidget: The project widget.
    """
    project_widget_frame = QtWidgets.QFrame()
    project_widget_ui = Ui_ProjectItemFrame()
    project_widget_ui.setupUi(project_widget_frame)
    project_widget_ui.projectNameLabel.setText(textwrap.fill(project.name or "", width=80, max_lines=1))
    project_widget_ui.projectNameLabel.setToolTip(project.name)
    project_widget_ui.modifiedDateTimeLabel.setText(
      datetime.datetime.fromisoformat(project.date or "").strftime("%Y-%m-%d %H:%M:%S"))
    return project_widget_frame

  def start_upload(self) -> None:
    """
    Starts the upload process.

    Explanation:
        This method starts the upload process for the selected projects.
        It retrieves the selected projects, creates upload widgets, and adds them to the upload queue.
        It also starts the upload manager task to process the upload queue.

    """
    for widget_pos in range(self.projectsScrollAreaVerticalLayout.count()):
      project_widget = self.projectsScrollAreaVerticalLayout.itemAt(widget_pos).widget()
      if project_widget.findChild(QtWidgets.QCheckBox, name="projectCheckBox").isChecked():
        upload_widget = self.get_upload_widget(
          project_widget.findChild(QtWidgets.QLabel, name="projectNameLabel").toolTip())
        self.uploadQueueVerticalLayout.addWidget(upload_widget["base"])
        task_thread = TaskThreadExtension(DataUploadTask(upload_widget["widget"]))
        self.upload_manager_task.add_to_queue(task_thread)
        if isinstance(task_thread.task, DataUploadTask):
          task_thread.task.uploadModelCreated.connect(upload_widget["widget"].modelIdLabel.setText)
    self.upload_manager_task_thread.task.start.emit()

  def clear_finished(self) -> None:
    """
    Clears the finished upload widgets.

    Explanation:
        This method clears the finished upload widgets from the upload queue.

    """
    for widget_pos in reversed(range(self.uploadQueueVerticalLayout.count())):
      self.uploadQueueVerticalLayout.itemAt(widget_pos).widget().setParent(None)

  def select_deselect_all_projects(self, checked: bool) -> None:
    """
    Selects or deselects all projects.

    Explanation:
        This method selects or deselects all projects based on the provided checked value.
        It iterates over the project widgets and sets the checked state of the projectCheckBox.

    Args:
        checked (bool): The checked state to set for the projectCheckBox.

    """
    for widget_pos in range(self.projectsScrollAreaVerticalLayout.count()):
      project_widget = self.projectsScrollAreaVerticalLayout.itemAt(widget_pos).widget()
      project_widget.findChild(QtWidgets.QCheckBox, name="projectCheckBox").setChecked(checked)

  def show_configure_upload(self) -> None:
    """
    Shows the configure upload dialog.

    Explanation:
        This method shows the configure upload dialog.

    """
    self.config_upload_dialog.load_ui()
    self.config_upload_dialog.instance.show()

  def show_completed_uploads(self) -> None:
    """
    Shows the completed uploads dialog.

    Explanation:
        This method shows the completed uploads dialog.

    """
    self.completed_uploads_dialog.load_ui()
    self.completed_uploads_dialog.instance.show()

  def show_edit_metadata(self) -> None:
    """
    Shows the edit metadata dialog.

    Explanation:
        This method shows the edit metadata dialog.

    """
    self.edit_metadata_dialog.show()

  def release_upload_manager(self) -> None:
    """
    Releases the upload manager.

    Explanation:
        This method releases the upload manager by quitting the upload manager task thread.

    """
    self.upload_manager_task_thread.quit()

  def close_ui(self) -> None:
    """
    Closes the UI.

    Explanation:
        This method closes the UI by quitting the upload manager task thread and adding a delay of 0.5 seconds.

    """
    self.upload_manager_task_thread.quit()
    time.sleep(0.5)

  def show_hide_log(self, button: QtWidgets.QPushButton) -> None:
    """
    Shows or hides the log console.

    Explanation:
        This method shows or hides the log console based on the current state.
        It retrieves the log console text edit widget and toggles its visibility.
        It also retrieves the model ID label and updates the log console text based on the model ID.

    Args:
        button (QtWidgets.QPushButton): The button triggering the action.

    """
    frame = button.parent()
    log_console_text_edit = frame.findChild(QtWidgets.QTextEdit, name="logConsoleTextEdit")
    if isinstance(log_console_text_edit, QtWidgets.QTextEdit):
      if log_console_text_edit.isHidden():
        log_console_text_edit.show()
      else:
        log_console_text_edit.hide()
      id_label = frame.findChild(QtWidgets.QLabel, name="modelIdLabel")
      if upload_model_id := id_label.text() if isinstance(id_label, QtWidgets.QLabel) else "":
        model = self.db_api.get_model(upload_model_id, UploadModel)
        log_console_text_edit.setText(model.log if isinstance(model, UploadModel) else "")


if __name__ == "__main__":
  import sys

  app = QtWidgets.QApplication(sys.argv)

  ui = MainDialog()
  ui.instance.show()
  sys.exit(app.exec())



""" ReorderColumnDelegate for the table views """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2023
#
#  Author: Jithu Murugan
#  Filename: reorder_column_delegate.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.

import logging
from typing import Union

from PySide6.QtCore import QAbstractItemModel, QEvent, QModelIndex, QPersistentModelIndex, QSize, Signal
from PySide6.QtGui import QPainter
from PySide6.QtWidgets import QApplication, QPushButton, QStyle, QStyleOptionButton, QStyleOptionViewItem, \
  QStyledItemDelegate, QWidget

from .utility_functions import is_click_within_bounds


class ReorderColumnDelegate(QStyledItemDelegate):
  """
  Delegate for creating the icons for the re-order column in the data hierarchy editor tables
  """
  re_order_signal = Signal(int)

  def __init__(self) -> None:
    """
      Constructor
    """
    super().__init__()
    self.logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")

  def paint(self,
            painter: QPainter,
            option: QStyleOptionViewItem,
            index: Union[QModelIndex, QPersistentModelIndex]) -> None:
    """
    Draws the re-order button within the cell represented by index
    Args:
      painter (QPainter): Painter instance for painting the button.
      option (QStyleOptionViewItem): Style option for the cell represented by index.
      index (Union[QModelIndex, QPersistentModelIndex]): Table cell index

    Returns: None

    """
    button = QPushButton()
    opt = QStyleOptionButton()
    opt.state = QStyle.State_Active | QStyle.State_Enabled  # type: ignore[operator]
    opt.rect = option.rect
    opt.icon = QApplication.style().standardIcon(QStyle.StandardPixmap.SP_ArrowUp)
    opt.iconSize = QSize(15, 15)
    QApplication.style().drawControl(QStyle.CE_PushButton, opt, painter, button)

  def createEditor(self,
                   parent: QWidget,
                   option: QStyleOptionViewItem,
                   index: Union[QModelIndex, QPersistentModelIndex]) -> QWidget:
    """
    Disable the editor for the whole re-order column by simply returning None
    Args:
      parent (QWidget): Parent table view.
      option (QStyleOptionViewItem): Style option for the cell represented by index.
      index (Union[QModelIndex, QPersistentModelIndex]): Cell index.

    Returns: None

    """
    return None  # type: ignore[return-value]

  def editorEvent(self,
                  event: QEvent,
                  model: QAbstractItemModel,
                  option: QStyleOptionViewItem,
                  index: Union[QModelIndex, QPersistentModelIndex]) -> bool:
    """
    In case of mouse click event, the re_order_signal is emitted for the respective table cell position
    Args:
      event (QEvent): The editor event information.
      model (QAbstractItemModel): Model data representing the table view.
      option (QStyleOptionViewItem): QStyleOption for the table cell.
      index (Union[QModelIndex, QPersistentModelIndex]): Table cell index.

    Returns (bool): True/False

    """
    if is_click_within_bounds(event, option):
      row = index.row()
      self.logger.info("Re-order signal emitted for the position: {%s} in the table..", row)
      self.re_order_signal.emit(row)
      return True
    return False



""" Represents an upload model object. """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: upload_model.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.
from pasta_eln.dataverse.base_model import BaseModel
from pasta_eln.dataverse.incorrect_parameter_error import IncorrectParameterError


class UploadModel(BaseModel):
  """
  Represents an upload model object.

  This class inherits from the BaseModel class and provides additional properties and methods for managing upload-related information.

  Args:
      _id (str | None): The ID of the object. Defaults to an empty string.
      _rev (str | None): The revision of the object. Defaults to an empty string.
      data_type (str | None): The data type of the upload. Defaults to an empty string.
      project_name (str | None): The name of the project. Defaults to an empty string.
      status (str | None): The status of the upload. Defaults to an empty string.
      finished_date_time (str | None): The date and time when the upload finished. Defaults to an empty string.
      log (str): The log associated with the upload. Defaults to an empty string.
      dataverse_url (str | None): The URL of the Dataverse. Defaults to an empty string.

  Raises:
      IncorrectParameterError: If any of the parameters are not of type str or None.

  Attributes:
      data_type (str | None): The data type of the upload.
      project_name (str | None): The name of the project.
      project_doc_id (str | None): The document ID of the project.
      status (str | None): The status of the upload.
      finished_date_time (str | None): The date and time when the upload finished.
      log (str): The log associated with the upload.
      dataverse_url (str | None): The URL of the Dataverse.

  """

  def __init__(self,
               _id: str | None = None,
               _rev: str | None = None,
               data_type: str | None = None,
               project_name: str | None = None,
               project_doc_id: str | None = None,
               status: str | None = None,
               finished_date_time: str | None = None,
               log: str = "",
               dataverse_url: str | None = None):
    """
    Initializes an upload model object.

    Args:
        _id (str | None): The ID of the object. Defaults to None.
        _rev (str | None): The revision of the object. Defaults to None.
        data_type (str | None): The data type of the upload. Defaults to None.
        project_name (str | None): The name of the project. Defaults to None.
        project_doc_id (str | None): The document ID of the project. Defaults to None.
        status (str | None): The status of the upload. Defaults to None.
        finished_date_time (str | None): The date and time when the upload finished. Defaults to None.
        log (str): The log associated with the upload. Defaults to empty string.
        dataverse_url (str | None): The URL of the Dataverse. Defaults to None.

    Raises:
        IncorrectParameterError: If any of the parameters are not of type str or None.

    """

    super().__init__(_id, _rev)
    if isinstance(data_type, str | None):
      self._data_type: str | None = 'dataverse_upload' if data_type is None else data_type
    else:
      raise IncorrectParameterError(f"Expected string type for data_type but got {type(data_type)}")
    if isinstance(project_name, str | None):
      self._project_name: str | None = project_name
    else:
      raise IncorrectParameterError(f"Expected string type for project_name but got {type(project_name)}")
    if isinstance(project_doc_id, str | None):
      self._project_doc_id: str | None = project_doc_id
    else:
      raise IncorrectParameterError(f"Expected string type for project_doc_id but got {type(project_doc_id)}")
    if isinstance(status, str | None):
      self._status: str | None = status
    else:
      raise IncorrectParameterError(f"Expected string type for status but got {type(status)}")
    if isinstance(finished_date_time, str | None):
      self._finished_date_time: str | None = finished_date_time
    else:
      raise IncorrectParameterError(f"Expected string type for finished_date_time but got {type(finished_date_time)}")
    if isinstance(log, str):
      self._log: str = log
    else:
      raise IncorrectParameterError(f"Expected string type for log but got {type(log)}")
    if isinstance(dataverse_url, str | None):
      self._dataverse_url: str | None = dataverse_url
    else:
      raise IncorrectParameterError(f"Expected string type for dataverse_url but got {type(dataverse_url)}")

  @property
  def data_type(self) -> str | None:
    """
    Returns the data type of the upload.

    Returns:
        str | None: The data type of the upload.

    """
    return self._data_type

  @data_type.setter
  def data_type(self, value: str | None) -> None:
    """
    Sets the data type of the upload.

    Args:
        value (str | None): The data type value to be set.

    Raises:
        IncorrectParameterError: If the value is not of type str.

    """
    if isinstance(value, str | None):
      self._data_type = value
    else:
      raise IncorrectParameterError(f"Expected string type for data_type but got {type(value)}")

  @data_type.deleter
  def data_type(self) -> None:
    """
    Deletes the data type of the upload.

    """
    del self._data_type

  @property
  def project_doc_id(self) -> str | None:
    """
    Returns the project document ID associated with the upload.

    Returns:
        str | None: The project doc ID.

    """
    return self._project_doc_id

  @project_doc_id.setter
  def project_doc_id(self, value: str | None) -> None:
    """
    Sets the project document ID associated with the upload.

    Args:
        value (str | None): The project doc ID value to be set.

    Raises:
        IncorrectParameterError: If the value is not of type str.

    """
    if isinstance(value, str | None):
      self._project_doc_id = value
    else:
      raise IncorrectParameterError(f"Expected string type for project_doc_id but got {type(value)}")

  @project_doc_id.deleter
  def project_doc_id(self) -> None:
    """
    Deletes the project_doc_id attribute.

    Explanation:
        This method deletes the project_doc_id attribute.

    Args:
        self: The instance of the class.

    """
    del self._project_doc_id

  @property
  def project_name(self) -> str | None:
    """
    Returns the project name associated with the upload.

    Returns:
        str | None: The project name.

    """
    return self._project_name

  @project_name.setter
  def project_name(self, value: str | None) -> None:
    """
    Sets the project name associated with the upload.

    Args:
        value (str | None): The project name value to be set.

    Raises:
        IncorrectParameterError: If the value is not of type str.

    """
    if isinstance(value, str | None):
      self._project_name = value
    else:
      raise IncorrectParameterError(f"Expected string type for project_name but got {type(value)}")

  @project_name.deleter
  def project_name(self) -> None:
    """
    Deletes the project_name attribute.

    Explanation:
        This method deletes the project_name attribute.

    Args:
        self: The instance of the class.

    """
    del self._project_name

  @property
  def status(self) -> str | None:
    """
    Returns the status of the upload.

    Returns:
        str | None: The status of the upload.

    """
    return self._status

  @status.setter
  def status(self, value: str | None) -> None:
    """
    Sets the status of the upload.

    Args:
        value (str | None): The status value to be set.

    Raises:
        IncorrectParameterError: If the value is not of type str.

    """
    if isinstance(value, str | None):
      self._status = value
    else:
      raise IncorrectParameterError(f"Expected string type for status but got {type(value)}")

  @status.deleter
  def status(self) -> None:
    """
    Deletes the status of the upload.

    """
    del self._status

  @property
  def finished_date_time(self) -> str | None:
    """
    Returns the finished date and time of the upload.

    Returns:
        str | None: The finished date and time of the upload.

    """
    return self._finished_date_time

  @finished_date_time.setter
  def finished_date_time(self, value: str | None) -> None:
    """
    Sets the finished date and time of the upload.

    Args:
        value (str | None): The finished date and time value to be set.

    Raises:
        IncorrectParameterError: If the value is not of type str.

    """
    if isinstance(value, str | None):
      self._finished_date_time = value
    else:
      raise IncorrectParameterError(f"Expected string type for finished_date_time but got {type(value)}")

  @finished_date_time.deleter
  def finished_date_time(self) -> None:
    """
    Deletes the finished date and time of the upload.

    """
    del self._finished_date_time

  @property
  def log(self) -> str:
    """
    Returns the log of the upload.

    Returns:
        str: The log of the upload.

    """
    return self._log

  @log.setter
  def log(self, value: str) -> None:
    """
    Appends a log message to the existing log associated with the upload.

    Args:
        value (str): The log message to be appended.

    Raises:
        IncorrectParameterError: If the value is not of type str.

    """
    if isinstance(value, str):
      self._log += f"{value}\n"
    else:
      raise IncorrectParameterError(f"Expected string type for log but got {type(value)}")

  @log.deleter
  def log(self) -> None:
    """
    Deletes the log of the upload.

    """
    del self._log

  @property
  def dataverse_url(self) -> str | None:
    """
    Returns the URL of the Dataverse.

    Returns:
        str | None: The URL of the Dataverse.

    """
    return self._dataverse_url

  @dataverse_url.setter
  def dataverse_url(self, value: str | None) -> None:
    """
    Sets the URL of the Dataverse.

    Args:
        value (str | None): The URL value to be set.

    Raises:
        IncorrectParameterError: If the value is not of type str.

    """
    if isinstance(value, str | None):
      self._dataverse_url = value
    else:
      raise IncorrectParameterError(f"Expected string type for dataverse_url but got {type(value)}")

  @dataverse_url.deleter
  def dataverse_url(self) -> None:
    """
    Deletes the URL of the Dataverse.

    """
    del self._dataverse_url



""" Entire config dialog (dialog is blocking the main-window, as opposed to create a new widget-window)"""
import platform
from PySide6.QtWidgets import QDialog, QVBoxLayout, QTabWidget  # pylint: disable=no-name-in-module
from ..guiCommunicate import Communicate
from .configGUI import ConfigurationGUI
from .configAuthors import ConfigurationAuthors
if platform.system()=='Windows':
  from .configSetupWindows import ConfigurationSetup
else:
  from .configSetupLinux import ConfigurationSetup

class Configuration(QDialog):
  """ Main class of entire config dialog """
  def __init__(self, comm:Communicate, startTab:str=''):
    """
    Initialization

    Args:
      backend (Pasta): backend, not communication
      startTab (str): tab to show initially
    """
    super().__init__()
    self.comm = comm
    self.setWindowTitle('PASTA-ELN configuration')

    # GUI elements
    mainL = QVBoxLayout(self)
    tabW = QTabWidget(self)
    mainL.addWidget(tabW)

    # Misc configuration: e.g. theming...
    tabGUI = ConfigurationGUI(self.comm, self.finished)
    tabW.addTab(tabGUI, 'Appearance')
    tabGUI = ConfigurationAuthors(self.comm, self.finished)
    tabW.addTab(tabGUI, 'Author')

    # Setup / Troubleshoot Pasta: main widget
    tabSetup = ConfigurationSetup(self.comm, self.finished)
    tabW.addTab(tabSetup, 'Setup')

    if startTab=='setup':
      tabW.setCurrentWidget(tabSetup)
      tabW.setTabEnabled(0, False)
      tabW.setTabEnabled(1, False)


  def finished(self, restart:bool=True) -> None:
    """
    callback function to close widget

    Args:
      restart (bool): if true, initialize
    """
    self.close()
    if restart:
      self.comm.backend.initialize()  #restart backend
    return



""" Provides an interface to interact with a database. """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: base_database_api.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.
import logging
from typing import Any

from cloudant import couchdb
from cloudant.design_document import DesignDocument
from cloudant.document import Document
from cloudant.error import CloudantDatabaseException
from cloudant.view import View

from pasta_eln.dataverse.config_error import ConfigError
from pasta_eln.dataverse.database_error import DatabaseError
from pasta_eln.dataverse.utils import log_and_create_error, read_pasta_config_file


class BaseDatabaseAPI:
  """
  Provides an interface to interact with a database.

  Explanation:
      This class encapsulates the functionality to interact with a database, including creating, retrieving, and updating documents,
      adding views to design documents, and getting view results.

  """

  def __init__(self) -> None:
    """
    Initializes the BaseDatabaseAPI instance.

    Explanation:
        This method initializes the BaseDatabaseAPI instance by setting up the necessary attributes and loading the configuration file.

    Args:
        None

    Raises:
        DatabaseError: If the config file or required fields are not found.

    Returns:
        None
    """
    super().__init__()
    self.password = None
    self.username = None
    self.host = 'localhost'
    self.port = 5984
    self.url = f'http://{self.host}:{self.port}'
    self.logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")
    config = read_pasta_config_file(self.logger)
    if 'defaultProjectGroup' in config:
      self.db_name = config['defaultProjectGroup']
    else:
      raise log_and_create_error(self.logger, ConfigError, "Incorrect config file, defaultProjectGroup not found!")
    self.set_username_password(config)

  def set_username_password(self, config: dict[str, Any]) -> None:
    """
    Sets the username and password based on the provided config.

    Args:
        config (dict[str, Any]): The configuration dictionary.

    Raises:
        DatabaseError: If the config file or required fields are not found.

    """
    project_groups = config.get('projectGroups')
    if not project_groups:
      raise log_and_create_error(self.logger, ConfigError,
                                 "Incorrect config file, projectGroups not found!")
    main_group = project_groups.get(self.db_name)
    if not main_group:
      raise log_and_create_error(self.logger, ConfigError,
                                 "Incorrect config file, defaultProjectGroup not found!")
    local_info = main_group.get('local')
    if not local_info:
      raise log_and_create_error(self.logger, ConfigError,
                                 "Incorrect config file, user or password not found!")
    self.username = local_info.get('user')
    self.password = local_info.get('password')

  def create_document(self, data: dict[str, Any]) -> Document:
    """
    Creates a document in the database with the provided data.

    Args:
        data (dict[str, Any]): The data to be stored in the document.

    Returns:
        Document: The created document.

    """
    self.logger.info("Creating document with data: %s", data)
    with couchdb(self.username,
                 self.password,
                 url=self.url,
                 connect=True) as client:
      pasta_db = client[self.db_name]
      document = None
      try:
        document = pasta_db.create_document(data, throw_on_exists=True)
      except CloudantDatabaseException as e:
        self.logger.error("Error creating document: %s", e)
      return document

  def get_document(self, document_id: str) -> Document:
    """
    Retrieves a document from the database based on the provided document ID.

    Args:
        document_id (str): The ID of the document to retrieve.

    Returns:
        Document: The retrieved document.

    Raises:
        DatabaseError: If the document ID is empty or if an error occurs during retrieval.

    """
    self.logger.info("Retrieving document with id: %s", document_id)
    if not document_id:
      raise log_and_create_error(self.logger, DatabaseError, "Document ID cannot be empty!")
    with couchdb(self.username,
                 self.password,
                 url=self.url,
                 connect=True) as client:
      pasta_db = client[self.db_name]
      document = None
      try:
        document = pasta_db[document_id]
      except KeyError as e:
        self.logger.error("Error retrieving document: %s", e)
      return document

  def update_document(self, data: dict[str, Any]) -> None:
    """
    Updates a document in the database with the provided data.
    Make sure to use the `_id` and `_rev` keys in the data to identify the document to update otherwise a new document will be created.

    Args:
        data (dict[str, Any]): The data to update the document with.

    """
    self.logger.info("Updating document with data: %s", data)
    with couchdb(self.username,
                 self.password,
                 url=self.url,
                 connect=True) as client:
      pasta_db = client[self.db_name]
      with Document(pasta_db, data['_id']) as document:
        for key, value in data.items():
          if key not in ['_id', '_rev']:
            document[key] = value

  def add_view(self,
               design_document_name: str,
               view_name: str,
               map_func: str,
               reduce_func: str | None = None) -> None:
    """
    Adds a view to the specified design document in the database.

    Args:
        design_document_name (str): The name of the design document.
        view_name (str): The name of the view.
        map_func (str): The map function for the view.
        reduce_func (str, optional): The reduce function for the view. Defaults to None.

    Raises:
        ValueError: If design_document_name, view_name, or map_func is None.

    Examples:
        # Add a view with a map function
        add_view("my_design_doc", "my_view", "function(doc) { emit(doc._id, 1); }")

        # Add a view with a map and reduce function
        add_view("my_design_doc", "my_view", "function(doc) { emit(doc.name, doc.age); }", "_count")
    """
    self.logger.info("Adding view: %s to design document: %s, map_func: %s, reduce_func: %s",
                     view_name,
                     design_document_name,
                     map_func,
                     reduce_func)
    if design_document_name is None or view_name is None or map_func is None:
      raise log_and_create_error(self.logger, DatabaseError,
                                 "Design document name, view name, and map function cannot be empty!")
    with couchdb(self.username,
                 self.password,
                 url=self.url,
                 connect=True) as client:
      pasta_db = client[self.db_name]
      with DesignDocument(pasta_db, design_document_name) as design_doc:
        design_doc.add_view(view_name, map_func, reduce_func)

  def get_view(self,
               design_document_name: str,
               view_name: str) -> View:
    """
    Retrieves a view from the design document.

    Explanation:
        This method retrieves a view from the specified design document.
        It connects to the CouchDB instance, selects the appropriate database,
        and retrieves the view from the design document.

    Args:
        self: The DatabaseAPI instance.
        design_document_name (str): The name of the design document.
        view_name (str): The name of the view.

    Raises:
        ValueError: If the design document name or view name is empty.

    Returns:
        View: The retrieved view.
    """
    self.logger.info("Retrieving view: %s from design document: %s",
                     view_name,
                     design_document_name)
    if design_document_name is None or view_name is None:
      raise log_and_create_error(self.logger, DatabaseError, "Design document name, view name cannot be empty!")
    with couchdb(self.username,
                 self.password,
                 url=self.url,
                 connect=True) as client:
      pasta_db = client[self.db_name]
      with DesignDocument(pasta_db, design_document_name) as design_doc:
        return design_doc.get_view(view_name)

  def get_view_results(self,
                       design_document_name: str,
                       view_name: str,
                       map_func: str | None = None,
                       reduce_func: str | None = None) -> list[dict[str, Any]]:
    """
    Gets the results of a view from the specified design document in the database.

    Args:
        design_document_name (str): The name of the design document.
        view_name (str): The name of the view.
        map_func (str, optional): The map function for the view. Defaults to None.
        reduce_func (str, optional): The reduce function for the view. Defaults to None.

    Raises:
        ValueError: If design_document_name or view_name is None.

    Returns:
        list[dict[str, Any]]: The results of the view as a list of dictionaries.

    Examples:
        # Get results from a view without map and reduce functions
        results = get_view_results("my_design_doc", "my_view")

        # Get results from a view with map and reduce functions
        results = get_view_results("my_design_doc", "my_view", "function(doc) { emit(doc._id, 1); }", "_count")
    """
    self.logger.info("Getting view results: %s from design document: %s, map_func: %s, reduce_func: %s",
                     view_name,
                     design_document_name,
                     map_func,
                     reduce_func)
    if design_document_name is None or view_name is None:
      raise log_and_create_error(self.logger, DatabaseError, "Design document name and view name cannot be empty!")
    with couchdb(self.username,
                 self.password,
                 url=self.url,
                 connect=True) as client:
      pasta_db = client[self.db_name]
      results: list[dict[str, Any]] = []
      with DesignDocument(pasta_db, design_document_name) as design_doc:
        view = View(design_doc, view_name, map_func, reduce_func)
        results.extend(result['value'] for result in view.result)
        return results

  def get_view_results_by_id(self,
                             design_document_name: str,
                             view_name: str,
                             document_id: str,
                             map_func: str | None = None,
                             reduce_func: str | None = None) -> Document | None:
    """
    Gets the result of a view for a specific document ID from the specified design document in the database.

    Args:
        design_document_name (str): The name of the design document.
        view_name (str): The name of the view.
        document_id (str): The ID of the document.
        map_func (str | None, optional): The map function for the view. Defaults to None.
        reduce_func (str | None, optional): The reduce function for the view. Defaults to None.

    Raises:
        ValueError: If design_document_name, view_name, or document_id is None.

    Returns:
        Document | None: The result of the view for the specified document ID, or None if not found.

    Examples:
        # Get the result of a view for a specific document ID
        result = get_view_results_by_id("my_design_doc", "my_view", "my_document_id")

        # Get the result of a view for a specific document ID with map and reduce functions
        result = get_view_results_by_id("my_design_doc", "my_view", "my_document_id", "function(doc) { emit(doc._id, 1); }", "_count")
    """
    self.logger.info("Getting view result: %s for id: %s from design document: %s, map_func: %s, reduce_func: %s",
                     view_name,
                     document_id,
                     design_document_name,
                     map_func,
                     reduce_func)
    with couchdb(self.username,
                 self.password,
                 url=self.url,
                 connect=True) as client:
      if design_document_name is None or view_name is None or document_id is None:
        raise log_and_create_error(self.logger, DatabaseError,
                                   "Design document name, view name and document id cannot be empty!")
      pasta_db = client[self.db_name]
      with DesignDocument(pasta_db, design_document_name) as design_doc:
        view = View(design_doc, view_name, map_func, reduce_func)
        return view.result[document_id][0]['value']



'''  Methods that check, repair, the local PASTA-ELN installation: no Qt-here '''
import os, platform, sys, json, shutil, random, string, logging, uuid
from typing import Optional, Any, Callable
import urllib.request
from pathlib import Path
from cloudant.client import CouchDB

from .backend import Backend
from .fixedStringsJson import defaultDataHierarchy, defaultConfiguration, configurationGUI
from .miscTools import outputString, DummyProgressBar


def getOS() -> str:
  '''
  Get operating system and python environment

  Returns:
    string: os + pythonEnvironment
  '''
  operatingSys = platform.system()
  # Get base/real prefix, or sys.prefix if there is none
  get_base_prefix_compat = getattr(sys, 'base_prefix', None) or getattr(sys, 'real_prefix', None) or sys.prefix
  in_virtualenv = get_base_prefix_compat != sys.prefix
  environment = sys.prefix if in_virtualenv else '_system_'
  return f'{operatingSys} {environment}'


def createDefaultConfiguration(user:str, password:str, pathPasta:Optional[Path]=None) -> dict[str,Any]:
  '''
  Create base of configuration file .pastaELN.json
  - basic project group
  - defaultProjectGroup
  - userID

  Args:
    user (str): user name (for windows)
    password (str): password (for windows)
    pathPasta (Path): place to store pasta data

  Returns:
    dict: dictionary of configuration
  '''
  if not user:
    user = input('Enter user name: ')
  if not password:
    password = input('Enter password: ')
  if pathPasta is None:
    if platform.system()=='Windows':
      pathPasta = Path.home()/'Documents'/'PASTA_ELN'
    else:
      pathPasta = Path.home()/'PASTA_ELN'
  conf: dict[str, Any] = {
      'defaultProjectGroup': 'research',
      'projectGroups': {
          'research': {
              'local': {'user': user, 'password': password, 'database': 'research', 'path': str(pathPasta)},
              'remote': {},
          }}, 'version': 2}
  try:
    conf['userID']      = os.getlogin()
  except Exception:   #github action
    conf['userID']      = 'github_user'
  #create pastaDir if it does not exist
  if not pathPasta.is_dir():
    pathPasta.mkdir()
  return conf


def runAsAdminWindows(cmdLine:list[str]) -> None:
  '''
  Run a command as admin in windows
  - (C) COPYRIGHT © Preston Landers 2010
  - https://stackoverflow.com/questions/19672352/how-to-run-script-with-elevated-privilege-on-windows
  - asks user for approval
  - waits for commands to end

  Args:
    cmdLine (list): list of command line sections runAsAdmin(["c:\\Windows\\notepad.exe"])
  '''
  import win32con, win32event, win32process
  from win32com.shell.shell import ShellExecuteEx
  from win32com.shell import shellcon
  procInfo = ShellExecuteEx(nShow=win32con.SW_SHOWNORMAL,
                            fMask=shellcon.SEE_MASK_NOCLOSEPROCESS,
                            lpVerb='runas',  # causes UAC elevation prompt.
                            lpFile= f'"{cmdLine[0]}"',
                            lpParameters=" ".join(cmdLine[1:]))
  procHandle = procInfo['hProcess']
  _ = win32event.WaitForSingleObject(procHandle, win32event.INFINITE)
  _   = win32process.GetExitCodeProcess(procHandle)
  return


def checkForDotNetVersion() -> bool:
  ''' check if .NET version 3.5 is installed (only on windows)
  '''
  import subprocess
  command = ['reg','query','HKEY_LOCAL_MACHINE\\SOFTWARE\\Microsoft\\NET Framework Setup\\NDP','/s']
  result = subprocess.run(command, stdout=subprocess.PIPE, check=True)
  #### For reference: code to show all installed versions
  # lines = [str(i).strip() for i in result.stdout.split(b'\n')]
  # lines = [i.split()[3][:-3] for i in lines if 'Version' in i]
  # versions = set(lines)
  # count3_5 = len([i for i in versions if i.startswith('3.5.')])
  # return count3_5>0
  return b'Version    REG_SZ    3.5' in result.stdout


def couchdb(command:str='test') -> str:
  '''
  test couchDB installation or (install it on Windows-only)

  Args:
    command (string): 'test' or 'install'

  Returns:
    string: '' for success, filled with errors
  '''
  if command == 'test':
    try:
      with urllib.request.urlopen('http://127.0.0.1:5984') as package:
        contents = package.read()
        if json.loads(contents)['couchdb'] == 'Welcome':
          return ''
    except Exception:
      pass
    return '**ERROR**'

  elif command == 'install':
    if platform.system()=='Linux':
      return '**ERROR should not be called'
    if platform.system()=='Windows':
      logging.info('CouchDB starting ...')
      url = 'https://couchdb.neighbourhood.ie/downloads/3.1.1/win/apache-couchdb-3.1.1.msi'
      path = Path.home()/'Downloads'/'apache-couchdb-3.1.1.msi'
      logging.info('Start download of couchdb')
      _, _ = urllib.request.urlretrieve(url, path)
      ## Old version with installer
      # cmd = ['cmd.exe','/K ',str(resultFilePath)]
      # _ = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, check=True)
      ## New version without questions
      password = ''.join(random.choice(string.ascii_letters) for _ in range(12))
      logging.info('PASSWORD: %s',password)
      pathS = str(path).replace('\\','\\\\')
      cmd = ['msiexec','/i',pathS,'/quiet','COOKIEVALUE=abcdefghijklmo','INSTALLSERVICE=1','ADMINUSER=admin',\
             f'ADMINPASSWORD={password}','/norestart','/l*','log.txt']
      logging.info('COMMAND: '+' '.join(cmd))
      runAsAdminWindows(cmd)
      logging.info('CouchDB ending')
      return f'Installed couchDB with password |{password}|'
    return f'**ERROR: Unknown operating system {platform.system()}'
  return '**ERROR: Unknown command'


def couchdbUserPassword(username:str, password:str) -> bool:
  '''
  test if username and password are correct

  Args:
    username (string): user name
    password (string): password

  Returns:
    bool: True if success, False if failure
  '''
  try:
    _ = CouchDB(username, password, url='http://127.0.0.1:5984', connect=True)
    return True
  except Exception:
    return False


def installLinuxRoot(couchDBExists:bool, pathPasta:Path=Path(''), password:str='') -> str:
  '''
  Install all packages in linux using the root-password

  Args:
    couchDBExists (bool): does the couchDB installation exist
    pathPasta (Path): path to install pasta in (Linux)
    password (str): password for couchDB installation

  Returns:
    string: ''=success, else error messages
  '''
  logging.info('InstallLinuxRoot starting ...')
  bashCommand = []
  password = ''
  if not couchDBExists:
    if not password:
      password = uuid.uuid4().hex
      logging.info('PASSWORD: %s',password)
    #create or adopt .pastaELN.json
    path = Path.home()/'.pastaELN.json'
    if path.is_file():
      with open(path,'r', encoding='utf-8') as fConf:
        conf = json.load(fConf)
      logging.info('.pastaELN.json exists, do not change it')
    else:
      conf = createDefaultConfiguration('admin', password, pathPasta)
      with open(path,'w', encoding='utf-8') as fConf:
        fConf.write(json.dumps(conf, indent=2) )
    bashCommand = [
      'sudo apt-get install -y libegl1',
      'sudo snap install couchdb',
      f'sudo snap set couchdb admin={password} setcookie={uuid.uuid4().hex}',
      'sudo snap start couchdb',
      'sudo snap connect couchdb:mount-observe',
      'sleep 5',
      f'curl -X PUT http://admin:{password}@127.0.0.1:5984/_users',
      f'curl -X PUT http://admin:{password}@127.0.0.1:5984/_replicator',
      f'curl -X PUT http://admin:{password}@127.0.0.1:5984/_global_changes',
      f'curl -X PUT http://admin:{password}@127.0.0.1:5984/_node/_local/_config/couch_httpd_auth/timeout/ -d \'"60000"\'',
      'sleep 10']
  #Try all terminals
  scriptFile = Path.home()/'pastaELN_Install.sh'
  with open(scriptFile,'w', encoding='utf-8') as shell:
    shell.write('\n'.join(bashCommand))
  os.chmod(scriptFile, 0o0777)
  terminals = ['xterm -e bash -c ','qterminal -e bash -c ','gnome-terminal -- ']
  logging.info('Command: %s',str(bashCommand))
  resultString = f'Password: {password}'
  for term in terminals:
    # _ = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, check=True)
    res = os.system(term+scriptFile.as_posix())
    logging.info('Linux install terminal %s  %s',term,str(res))
    if res == 0:
      break
    if terminals.index(term)==len(terminals)-1:
      logging.error('**ERROR: Last terminal failed')
      res = os.system('\n'.join(bashCommand[:-2]))
      logging.info('Finished using straight Bash command result=%s',str(res))
      resultString = '**ERROR: Last terminal failed'
  success = 'CouchDB works' if couchdbUserPassword('admin',password) else 'CouchDB FAILED'
  logging.info('InstallLinuxRoot ending. %s',success)
  return resultString


def configuration(command:str='test', user:str='', password:str='', pathPasta:Path=Path('')) -> str:
  '''
  Check configuration file .pastaELN.json for consistencies

  Args:
    command (str): 'test' or 'repair'
    user (str): user name (for windows)
    password (str): password (for windows)
    pathPasta (Path): path to install pasta in (Windows)

  Returns:
    string: ''=success, else error messages
  '''
  logging.info('Configuration starting ...')
  output = ''
  try:
    with open(Path.home()/'.pastaELN.json','r', encoding='utf-8') as fConf:
      conf = json.load(fConf)
  except Exception:
    output += '**ERROR configuration file does not exist\n'
    conf = {}
    if command == 'repair':
      conf = createDefaultConfiguration(user, password, pathPasta)
  logging.info(json.dumps(conf, indent=2))

  #check normal items
  for k,v in defaultConfiguration.items():
    if k not in conf:
      if command == 'repair':
        if isinstance(v,str) and v.startswith('$') and v.startswith('$'):
          v = eval(v[1:-1]) # pylint: disable=eval-used
        conf[k] = v
      else:
        output += outputString('text','error', f'No {k} in config file')
  #GUI items
  if 'GUI' not in conf:
    if command == 'repair':
      conf['GUI'] = {}
    else:
      output += outputString('text','error', 'No GUI in config file')
  for _, items in configurationGUI.items():
    for k,v in items.items():
      if 'GUI' in conf and k not in conf['GUI']:
        if command == 'repair':
          conf['GUI'][k] = v[1]
        else:
          output += outputString('text','error', f'No {k} in GUI part of config file')
  if command == 'repair':
    with open(Path.home()/'.pastaELN.json','w', encoding='utf-8') as f:
      f.write(json.dumps(conf,indent=2))
  logging.info('Configuration ending')
  return output



def dataHierarchy(command:str='test') -> str:
  '''
  Check configuration file .pastaELN.json for consistencies

  Args:
    command (string): 'test' or 'install'

  Returns:
    string: ''=success, else error messages
  '''
  backend = Backend()
  if command == 'test':
    if not hasattr(backend.db, 'db'):
      return '**ERROR: couchDB not initialized'
    output = 'database name: {backend.db.db.database_name}\nDesign documents\n'
    for item in backend.db.db.design_documents():
      numViews = len(item['doc']['views']) if 'views' in item['doc'] else 0
      output += '  '+item['id']+'   Num. of views:'+str(numViews)+'\n'
    try:
      _ = backend.db.getDoc('-dataHierarchy-')
      output += 'DataHierarchy exists on server'+'\n'
    except Exception:
      output += '**ERROR: DataHierarchy does NOT exist on server'+'\n'
    return output
  elif command == 'install':
    logging.info('dataHierarchy starting ...')
    doc = defaultDataHierarchy
    logging.info(str(doc))
    logging.info('dataHierarchy ending ...')
    # _ = backend.db.create_document(doc)
    return ''

  return '**ERROR: Unknown command'



def exampleData(force:bool=False, callbackPercent:Optional[Callable[[int],None]]=None) -> str:
  '''
  Create example data after installation

  Args:
    force (bool): force creation by removing content before creation
    callbackPercent (function): callback function given to exampleData, such that exampleData can report progress back
  '''
  logging.info('Start example data creation')
  outputFormat = 'print'
  if callbackPercent is not None:
    callbackPercent(0)
  if force:
    backend = Backend(initConfig=False)
    dirName = backend.basePath
    backend.exit(deleteDB=True)
    shutil.rmtree(dirName)
    os.makedirs(dirName)
  if callbackPercent is not None:
    callbackPercent(1)
  backend = Backend(initViews=True, initConfig=False)
  if callbackPercent is not None:
    callbackPercent(2)
  ### CREATE PROJECTS AND SHOW
  outputString(outputFormat,'h2','CREATE EXAMPLE PROJECT AND SHOW')
  backend.addData('x0', {'-name': 'PASTAs Example Project', 'objective': 'Test if everything is working as intended.', 'status': 'active', 'comment': '#Important Can be used as reference or deleted'})
  if callbackPercent is not None:
    callbackPercent(3)
  outputString(outputFormat,'info', backend.output('x0'))
  if callbackPercent is not None:
    callbackPercent(4)
  logging.info('Finished creating example project')

  ### TEST PROJECT PLANING
  outputString(outputFormat,'h2','TEST PROJECT PLANING')
  viewProj = backend.db.getView('viewDocType/x0')
  projID1  = [i['id'] for i in viewProj if 'PASTA' in i['value'][0]][0]
  if callbackPercent is not None:
    callbackPercent(5)
  backend.changeHierarchy(projID1)
  if callbackPercent is not None:
    callbackPercent(6)
  backend.addData('x1',    {'comment': 'This is hard! #TODO', '-name': 'This is an example task'})
  if callbackPercent is not None:
    callbackPercent(7)
  currentID = backend.addData('x1',    {'comment': 'This will take a long time. #WAIT', '-name': 'This is another example task'})
  if callbackPercent is not None:
    callbackPercent(8)
  backend.changeHierarchy(currentID)
  backend.addData('x2',    {'-name': 'This is an example subtask',     'comment': 'Random comment 1'})
  if callbackPercent is not None:
    callbackPercent(9)
  backend.addData('x2',    {'-name': 'This is another example subtask','comment': 'Random comment 2'})
  if callbackPercent is not None:
    callbackPercent(10)
  backend.changeHierarchy(None)
  semStepID = backend.addData('x1',    {'-name': 'Data files'})
  if callbackPercent is not None:
    callbackPercent(11)
  backend.changeHierarchy(semStepID)
  if backend.cwd is not None:
    semDirName = backend.basePath/backend.cwd
  backend.changeHierarchy(None)
  outputString(outputFormat,'info',backend.outputHierarchy())
  if callbackPercent is not None:
    callbackPercent(12)
  logging.info('Finished project planning')

  ### TEST PROCEDURES
  outputString(outputFormat,'h2','TEST PROCEDURES')
  sopDir = backend.basePath/'StandardOperatingProcedures'
  os.makedirs(sopDir, exist_ok=True)
  with open(sopDir/'Example_SOP.md','w', encoding='utf-8') as fOut:
    fOut.write('# Put sample in instrument\n# Do something\nDo not forget to\n- not do anything wrong\n- **USE BOLD LETTERS**\n')
  if callbackPercent is not None:
    callbackPercent(13)
  backend.addData('procedure', {'-name': 'StandardOperatingProcedures/Example_SOP.md', 'comment': '#v1'})
  if callbackPercent is not None:
    callbackPercent(14)
  outputString(outputFormat,'info',backend.output('procedure'))
  if callbackPercent is not None:
    callbackPercent(15)
  logging.info('Finished procedures creating')

  ### TEST SAMPLES
  outputString(outputFormat,'h2','TEST SAMPLES')
  backend.addData('sample',    {'-name': 'Example sample', 'chemistry': 'A2B2C3', 'qrCode': '13214124 99698708', 'comment': 'can be used as example or removed'})
  if callbackPercent is not None:
    callbackPercent(16)
  outputString(outputFormat,'info',backend.output('sample'))
  if callbackPercent is not None:
    callbackPercent(17)
  outputString(outputFormat,'info',backend.outputQR())
  if callbackPercent is not None:
    callbackPercent(18)
  logging.info('Finished samples creating')

  ###  TEST MEASUREMENTS AND SCANNING/CURATION
  outputString(outputFormat,'h2','TEST MEASUREMENTS AND SCANNING')
  shutil.copy(Path(__file__).parent/'Resources'/'ExampleMeasurements'/'simple.png', semDirName)
  shutil.copy(Path(__file__).parent/'Resources'/'ExampleMeasurements'/'simple.csv', semDirName)
  if callbackPercent is not None:
    callbackPercent(19)
  logging.info('Finished copy files')
  progressBar = DummyProgressBar()
  backend.scanProject(progressBar, projID1)
  logging.info('Finished scan tree')
  if callbackPercent is not None:
    callbackPercent(20)

  ### USE GLOBAL FILES
  outputString(outputFormat,'h2','USE GLOBAL FILES')
  backend.changeHierarchy(semStepID)
  if callbackPercent is not None:
    callbackPercent(21)
  backend.addData('measurement', {
    '-name': 'https://upload.wikimedia.org/wikipedia/commons/thumb/a/a4/Misc_pollen.jpg/315px-Misc_pollen.jpg',\
    'comment':'remote image from wikipedia. Used for testing and reference. Can be deleted.'})
  if callbackPercent is not None:
    callbackPercent(22)
  outputString(outputFormat,'info',backend.output('measurement'))
  if callbackPercent is not None:
    callbackPercent(23)
  logging.info('Finished global files additions')

  ### VERIFY DATABASE INTEGRITY
  outputString(outputFormat,'h2','VERIFY DATABASE INTEGRITY')
  outputString(outputFormat,'info',backend.checkDB(outputStyle='text'))
  outputString(outputFormat,'h2','DONE WITH VERIFY')
  if callbackPercent is not None:
    callbackPercent(24)
  logging.info('Finished checking database')
  return 'Finished checking database'


def createShortcut() -> None:
  """
  Create alias and shortcut icon depending on operating system
  """
  logging.info('Create shortcut starting')
  if platform.system()=='Linux':
    content ='[Desktop Entry]\nName=PASTA ELN\nComment=PASTA electronic labnotebook\n'
    if sys.prefix==sys.base_prefix:   #normal installation into user-space
      content+='Exec=pastaELN\n'
    else:                             #installation in a virtual environment
      logging.info('In virtual environment, create an alias')
      with open(Path.home()/'.bashrc','a', encoding='utf-8') as fOut:
        alias = f"alias pastaELN='{sys.prefix}/bin/python3 -m pasta_eln.gui'"
        logging.info(alias)
        fOut.write(f'{alias}\n')
      content += f'Exec={sys.prefix}/bin/python3 -m pasta_eln.gui\n'
    content+='Icon='+ (Path(__file__).parent/'Resources'/'Icons'/'favicon64.png').as_posix() + '\n'
    content+='Terminal=false\nType=Application\nCategories=Science;Application;\n'
    try:
      linkString = (Path.home()/'Desktop'/'pastaELN.desktop').as_posix()
      with open(linkString,'w', encoding='utf-8') as fOut:
        fOut.write(content)
        os.system(f'gio set {linkString} metadata::trusted true') #for ubuntu systems
        os.chmod(Path.home()/'Desktop'/'pastaELN.desktop', 0o775)
    except Exception:
      pass
    try:
      with open(Path.home()/'.local'/'share'/'applications'/'pastaELN.desktop','w', encoding='utf-8') as fOut:
        fOut.write(content)
        os.chmod(Path.home()/'.local'/'share'/'applications'/'pastaELN.desktop', 0o777)
    except Exception:
      pass
  elif platform.system()=='Windows':
    import winshell
    from win32com.client import Dispatch
    shell = Dispatch('WScript.Shell')
    shortcut = shell.CreateShortCut( os.path.join(winshell.desktop(), "pastaELN.lnk") )
    env = os.environ['CONDA_PREFIX'] if 'CONDA_PREFIX' in os.environ else ''  #full path of env
    if env:
      env = env.split('\\')[-1]  #just the env name
      user = os.getlogin()
      batContent = f"cmd.exe /c \"C:\\Users\\{user}\\anaconda3\\Scripts\\activate {env} && python -m pasta_eln.gui\""
      batLocation= f"C:\\Users\\{user}\\startPastaELN.bat"
      with open(batLocation, 'w', encoding='utf-8') as fBat:
        fBat.write(batContent)
      shortcut.Targetpath = batLocation
    else:
      shortcut.Targetpath = r"python -m pasta_eln.gui"
    shortcut.WorkingDirectory = str(Path.home())
    shortcut.IconLocation = str(Path(__file__).parent/'Resources'/'Icons'/'favicon64.ico')
    shortcut.save()
  logging.info('Create shortcut end')
  return


##############
# Main method for testing and installation without GUI
def main() -> None:
  ''' Main method and entry point for commands '''
  logPath = Path.home()/'pastaELN.log'
  #old versions of basicConfig do not know "encoding='utf-8'"
  logging.basicConfig(filename=logPath, level=logging.INFO, format='%(asctime)s|%(levelname)s:%(message)s',
                      datefmt='%m-%d %H:%M:%S')   #This logging is always info, since for installation only
  for package in ['urllib3', 'requests', 'asyncio', 'PIL', 'matplotlib.font_manager']:
    logging.getLogger(package).setLevel(logging.WARNING)
  logging.info('Start PASTA Install')
  print('---- Test PASTA-ELN installation----')
  print('--   if nothing reported: it is ok.')
  print('getOS        :', getOS())
  res = couchdb()
  existsCouchDB = res==''
  print('chouchDB     :', res)
  res = configuration()
  flagConfiguration = 'ERROR' in res
  print('configuration:', res)
  try:
    res = '\n'+dataHierarchy()
  except Exception:
    res = ' **ERROR**'
    #No new 'raise' so that it install-script continues its path
  flagDataHierarchy = 'ERROR' in res
  print(f'dataHierarchy     :{res}')

  print('Add "install" argument to install PASTA-ELN.')
  if len(sys.argv)>1 and 'install' in sys.argv:
    if platform.system()=='Linux':
      print('---- Create PASTA-ELN installation Linux ----')
      if not existsCouchDB:
        print('install with root credentials...')
        installLinuxRoot(existsCouchDB, Path.home()/'pastaELN')
      if flagConfiguration:
        print('repair  configuration:', configuration('repair'))
      if flagDataHierarchy and existsCouchDB:
        print('install dataHierarchy     :', dataHierarchy('install'))

    elif platform.system()=='Windows':
      print('---- Create PASTA-ELN installation Windows ----')
      if not existsCouchDB:
        print('install couchDB      :', couchdb('install'))
      if flagConfiguration:
        print('repair  configuration:', configuration('repair'))
      if flagDataHierarchy and existsCouchDB:
        print('install dataHierarchy     :', dataHierarchy('install'))

  print('Add "example" argument to create example data.')
  if len(sys.argv)>1 and 'example' in sys.argv:
    print('---- Create Example data ----')
    print('create example data  :', exampleData())

  logging.info('End PASTA Install')
  return


# called by python3 -m pasta_eln.installTools
if __name__ == '__main__':
  main()



""" Constants used in the data hierarchy configuration """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2023
#
#  Author: Jithu Murugan
#  Filename: constants.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.

# Data hierarchy table view constants
METADATA_TABLE_LIST_COLUMN_INDEX = 2
METADATA_TABLE_IRI_COLUMN_INDEX = 4
METADATA_TABLE_REQUIRED_COLUMN_INDEX = 5
METADATA_TABLE_DELETE_COLUMN_INDEX = 6
METADATA_TABLE_REORDER_COLUMN_INDEX = 7

ATTACHMENT_TABLE_DELETE_COLUMN_INDEX = 2
ATTACHMENT_TABLE_REORDER_COLUMN_INDEX = 3

# Help page URL
DATA_HIERARCHY_HELP_PAGE_URL = "https://pasta-eln.github.io/pasta-eln/data_hierarchy_configuration.html"



""" Represents a configuration model object. """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: config_model.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.
from typing import Any

from pasta_eln.dataverse.base_model import BaseModel
from pasta_eln.dataverse.incorrect_parameter_error import IncorrectParameterError


class ConfigModel(BaseModel):
  """
  Represents a configuration model object.

  This class inherits from the BaseModel class and provides additional properties and methods for managing configuration-related information.

  Args:
      _id (str | None): The ID of the object. Defaults to None.
      _rev (str | None): The revision of the object. Defaults to None.
      project_upload_items (dict[str, Any] | None): A dictionary containing upload items of the project. Defaults to None.
      parallel_uploads_count (int): The number of parallel uploads configured. Defaults to None.
      dataverse_login_info (dict[str, Any]): A dictionary containing Dataverse login information. Defaults to None.
      metadata (dict[str, Any]): A dictionary containing metadata for the dataset in Dataverse. Defaults to None.

  Raises:
      IncorrectParameterError: If any of the arguments have an invalid type.

  Attributes:
      project_upload_items (dict[str, Any] | None): The upload items of the PASTA project.
      parallel_uploads_count (int | None): The number of parallel uploads configured.
      dataverse_login_info (dict[str, Any] | None): The Dataverse login information.
      metadata (dict[str, Any] | None): The metadata for the dataset in Dataverse.

  """

  def __init__(self,
               _id: str | None = None,
               _rev: str | None = None,
               project_upload_items: dict[str, Any] | None = None,
               parallel_uploads_count: int | None = None,
               dataverse_login_info: dict[str, Any] | None = None,
               metadata: dict[str, Any] | None = None):
    """
    Initializes a config model object.

    Args:
        _id (str | None): The ID of the object. Defaults to None.
        _rev (str | None): The revision of the object. Defaults to None.
        project_upload_items (dict[str, Any] | None): A dictionary containing upload items of the project. Defaults to None.
        parallel_uploads_count (int): The number of parallel uploads configured. Defaults to None.
        dataverse_login_info (dict[str, Any]): A dictionary containing Dataverse login information. Defaults to None.
        metadata (dict[str, Any]): A dictionary containing metadata for the dataset in Dataverse. Defaults to None.

    Raises:
        IncorrectParameterError: If any of the parameters have an invalid type.

    """
    super().__init__(_id, _rev)
    if isinstance(project_upload_items, dict | None):
      self._project_upload_items: dict[str, Any] | None = project_upload_items
    else:
      raise IncorrectParameterError(f"Expected dictionary for upload_items but got {type(project_upload_items)}")
    if isinstance(parallel_uploads_count, int | None):
      self._parallel_uploads_count: int | None = parallel_uploads_count
    else:
      raise IncorrectParameterError(
        f"Expected integer for parallel_uploads_count but got {type(parallel_uploads_count)}")
    if isinstance(dataverse_login_info, dict | None):
      self._dataverse_login_info: dict[str, Any] | None = dataverse_login_info
    else:
      raise IncorrectParameterError(
        f"Expected dictionary for dataverse_login_info but got {type(dataverse_login_info)}")
    if isinstance(metadata, dict | None):
      self._metadata: dict[str, Any] | None = metadata
    else:
      raise IncorrectParameterError(f"Expected dictionary for metadata but got {type(metadata)}")

  @property
  def project_upload_items(self) -> dict[str, Any] | None:
    """
    Returns the upload items of the project.

    Returns:
        dict[str, Any] | None: A dictionary containing the upload items of the project.

    """
    return self._project_upload_items

  @project_upload_items.setter
  def project_upload_items(self, value: dict[str, Any] | None) -> None:
    """
    Sets the upload items of the project.

    Args:
        value (dict[str, Any] | None): The upload items to set. Defaults to None.

    Raises:
        IncorrectParameterError: If the value is not of type dict or None.

    """
    if isinstance(value, dict | None):
      self._project_upload_items = value
    else:
      raise IncorrectParameterError(f"Expected dictionary for upload_items but got {type(value)}")

  @project_upload_items.deleter
  def project_upload_items(self) -> None:
    """
    Deletes the upload items of the project.

    """
    del self._project_upload_items

  @property
  def parallel_uploads_count(self) -> int | None:
    """
    Returns the number of parallel uploads.

    Returns:
        int | None: The number of parallel uploads.

    """
    return self._parallel_uploads_count

  @parallel_uploads_count.setter
  def parallel_uploads_count(self, value: int | None) -> None:
    """
    Sets the number of parallel uploads.

    Args:
        value (int | None): The number of parallel uploads to set.

    Raises:
        IncorrectParameterError: If the value is not of type int.

    """
    if isinstance(value, int | None):
      self._parallel_uploads_count = value
    else:
      raise IncorrectParameterError(f"Expected integer for parallel_uploads_count but got {type(value)}")

  @parallel_uploads_count.deleter
  def parallel_uploads_count(self) -> None:
    """
    Deletes the number of parallel uploads.

    """
    del self._parallel_uploads_count

  @property
  def dataverse_login_info(self) -> dict[str, Any] | None:
    """
    Returns the dataverse login information.

    Returns:
        dict[str, Any] | None: The dataverse login information.

    """
    return self._dataverse_login_info

  @dataverse_login_info.setter
  def dataverse_login_info(self, value: dict[str, Any] | None) -> None:
    """
    Sets the dataverse login information.

    Args:
        value (dict[str, Any] | None): The dataverse login information to set.

    Raises:
        IncorrectParameterError: If the value is not of type dict.

    """
    if isinstance(value, dict | None):
      self._dataverse_login_info = value
    else:
      raise IncorrectParameterError(f"Expected dictionary for dataverse_login_info but got {type(value)}")

  @dataverse_login_info.deleter
  def dataverse_login_info(self) -> None:
    """
    Deletes the dataverse login information.

    """
    del self._dataverse_login_info

  @property
  def metadata(self) -> dict[str, Any] | None:
    """
    Returns the metadata.

    Returns:
        dict[str, Any] | None: The metadata.

    """
    return self._metadata

  @metadata.setter
  def metadata(self, value: dict[str, Any] | None) -> None:
    """
    Sets the metadata.

    Args:
        value (dict[str, Any] | None): The metadata to set.

    Raises:
        IncorrectParameterError: If the value is not of type dict.

    """
    if isinstance(value, dict | None):
      self._metadata = value
    else:
      raise IncorrectParameterError(f"Expected dictionary for metadata but got {type(value)}")

  @metadata.deleter
  def metadata(self) -> None:
    """
    Deletes the metadata.

    """
    del self._metadata



""" Custom exception class for null data hierarchy document """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2023
#
#  Author: Jithu Murugan
#  Filename: document_null_exception.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.

class DocumentNullException(Exception):
  """
  Custom exception class for null data hierarchy document
  """

  def __init__(self,
               message: str,
               detailed_errors: dict[str, str]):
    """
    Constructs DocumentNullException
    Args:
      message (str): Error message to be thrown
      detailed_errors (dict): Additional errors passed via exception
    """
    super().__init__(message)
    self.message = message
    self.detailed_errors = detailed_errors
